# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Shared helpers for Orca GTK user interfaces."""

# pylint: disable=no-member
# pylint: disable=too-many-arguments
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals

from __future__ import annotations

import os
from dataclasses import dataclass
from typing import TYPE_CHECKING

import gi

gi.require_version("Atk", "1.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Atk, Gdk, Gio, GLib, Gtk  # pylint: disable=no-name-in-module

from . import debug

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any


@dataclass
class AppearanceProviders:
    """CSS providers for conditional appearance settings."""

    hc: Gtk.CssProvider
    dark: Gtk.CssProvider
    shapes: Gtk.CssProvider | None = None


@dataclass(frozen=True)
class ListRowAction:
    """Action button metadata for custom list rows."""

    name: str
    label: str
    callback: Callable[[Gtk.Button], None]
    icon_name: str | None = None


_BASE_CSS = b"""
    decoration {
        border-radius: 15px;
    }
    @define-color orca_sidebar_bg shade(@theme_bg_color, 0.98);
    list.frame {
        border-color: alpha(@theme_fg_color, 0.15);
    }
    .orca-left-headerbar {
        background-color: @orca_sidebar_bg;
        background-image: none;
        border-bottom-width: 0;
    }
    .orca-panel-headerbar {
        background-color: @theme_bg_color;
        background-image: none;
        border-bottom-width: 0;
    }
    .orca-sidebar {
        background-color: @orca_sidebar_bg;
    }
    .orca-sidebar scrolledwindow,
    .orca-sidebar list {
        background-color: transparent;
    }
    .orca-sidebar list {
        padding: 6px 0;
    }
    .orca-sidebar list row {
        border-radius: 9px;
        min-height: 36px;
        padding: 0 8px;
        margin: 0 6px 2px;
    }
    .orca-sidebar list row:selected {
        background-color: alpha(@theme_fg_color, 0.10);
    }
    .orca-sidebar list row:selected,
    .orca-sidebar list row:selected label {
        color: @theme_fg_color;
    }
    .orca-sidebar list row:hover {
        background-color: alpha(@theme_fg_color, 0.07);
    }
    .orca-sidebar list row:selected:hover {
        background-color: alpha(@theme_fg_color, 0.13);
    }
    .orca-sidebar list row:active,
    .orca-sidebar list row:selected:active {
        background-color: alpha(@theme_fg_color, 0.19);
    }
    list.frame row:focus {
        box-shadow: inset 0 0 0 2px alpha(@theme_selected_bg_color, 0.5);
    }
"""

_HIGH_CONTRAST_CSS = b"""
    list.frame {
        border-color: alpha(@theme_fg_color, 0.4);
    }
    list separator {
        background-color: alpha(@theme_fg_color, 0.4);
    }
    .dim-label {
        opacity: 1.0;
    }
"""

_DARK_MODE_CSS = b"""
    @define-color orca_sidebar_bg @theme_bg_color;
    window.background {
        background-color: @theme_base_color;
    }
    .orca-dialog {
        background-color: @theme_base_color;
    }
    .orca-dialog-content {
        background-color: @theme_base_color;
    }
    .orca-dialog-content list.frame,
    .orca-dialog-content scrolledwindow.frame {
        border-color: alpha(@theme_fg_color, 0.35);
    }
    .orca-panel-headerbar {
        background-color: @theme_base_color;
    }
    switch slider {
        background-image: image(white);
    }
"""

_GNOME_DARK_CSS = b"""
    @define-color theme_bg_color #303030;
    @define-color theme_base_color #242424;
    @define-color orca_sidebar_bg #2e2e32;
"""

_STATUS_SHAPES_CSS = b"""
    switch image {
        color: @theme_fg_color;
    }
    switch:checked image {
        color: white;
    }
"""

_appearance_refs: list[tuple] = []


def _style_provider_priority() -> int | None:
    """Return GTK's application CSS priority, if available."""

    priority = Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    if isinstance(priority, int):
        return priority
    return None


def sync_appearance() -> tuple | None:
    """Bridge GNOME color-scheme and high-contrast gsettings to GTK3."""

    if _appearance_refs:
        return _appearance_refs[0]

    gtk_settings = Gtk.Settings.get_default()  # pylint: disable=no-value-for-parameter
    screen = Gdk.Screen.get_default()  # pylint: disable=no-value-for-parameter
    if gtk_settings is None or screen is None:
        return None

    priority = _style_provider_priority()
    if priority is None:
        return None

    base_provider = Gtk.CssProvider()
    base_provider.load_from_data(_BASE_CSS)
    Gtk.StyleContext.add_provider_for_screen(
        screen,
        base_provider,
        priority,
    )

    providers = AppearanceProviders(
        hc=Gtk.CssProvider(),
        dark=Gtk.CssProvider(),
    )
    providers.hc.load_from_data(_HIGH_CONTRAST_CSS)
    dark_css = _DARK_MODE_CSS
    if "GNOME" in os.environ.get("XDG_CURRENT_DESKTOP", ""):
        dark_css += _GNOME_DARK_CSS
    providers.dark.load_from_data(dark_css)

    try:
        interface_settings = Gio.Settings(schema_id="org.gnome.desktop.interface")
        a11y_settings = Gio.Settings(schema_id="org.gnome.desktop.a11y.interface")
        if "show-status-shapes" in a11y_settings.list_keys():
            shapes_provider = Gtk.CssProvider()
            shapes_provider.load_from_data(_STATUS_SHAPES_CSS)
            providers.shapes = shapes_provider
        apply_appearance(
            interface_settings,
            a11y_settings,
            gtk_settings,
            screen,
            providers,
        )

        def on_setting_changed(*_args):
            apply_appearance(
                interface_settings,
                a11y_settings,
                gtk_settings,
                screen,
                providers,
            )

        interface_settings.connect("changed::color-scheme", on_setting_changed)
        a11y_settings.connect("changed::high-contrast", on_setting_changed)
        if providers.shapes is not None:
            a11y_settings.connect("changed::show-status-shapes", on_setting_changed)
        gtk_settings.connect("notify::gtk-theme-name", on_setting_changed)
        _appearance_refs.append((interface_settings, a11y_settings, base_provider, providers))
        return _appearance_refs[0]
    except GLib.Error as error:
        msg = f"ORCA GUI: Exception syncing appearance: {error}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
    return None


def create_headerbar(title: str) -> Gtk.HeaderBar:
    """Return an Orca-styled header bar."""

    headerbar = Gtk.HeaderBar()
    headerbar.set_show_close_button(True)
    headerbar.set_title(title)
    headerbar.get_style_context().add_class("orca-panel-headerbar")
    return headerbar


def set_margins(
    widget: Gtk.Widget,
    *,
    start: int = 0,
    end: int = 0,
    top: int = 0,
    bottom: int = 0,
) -> None:
    """Set all margins on a widget at once."""

    widget.set_margin_start(start)
    widget.set_margin_end(end)
    widget.set_margin_top(top)
    widget.set_margin_bottom(bottom)


def create_row_box(spacing: int = 12, margin: int = 12) -> Gtk.Box:
    """Return a standard horizontal row box."""

    hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=spacing)
    set_margins(hbox, start=margin, end=margin, top=margin, bottom=margin)
    return hbox


def create_vertical_row_box(spacing: int = 6, margin: int = 12) -> Gtk.Box:
    """Return a standard vertical row box."""

    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=spacing)
    set_margins(vbox, start=margin, end=margin, top=margin, bottom=margin)
    return vbox


def create_heading_label(text: str, margin_top: int = 12) -> Gtk.Label:
    """Return a heading label."""

    label = Gtk.Label(label=text, xalign=0)
    label.get_style_context().add_class("heading")
    label.set_margin_top(margin_top)
    label.set_margin_bottom(6)
    return label


def create_heading_action_box(
    heading: str,
    icon_name: str,
    accessible_name: str,
    clicked_handler: Callable[[Gtk.Button], None],
) -> tuple[Gtk.Box, Gtk.Button]:
    """Return a heading row with a trailing icon button."""

    header_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
    header_box.set_margin_bottom(6)

    title_label = create_heading_label(heading, margin_top=0)
    title_label.set_margin_bottom(0)
    title_label.set_halign(Gtk.Align.START)
    header_box.pack_start(title_label, True, True, 0)

    button = Gtk.Button.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
    button.get_accessible().set_name(accessible_name)
    button.connect("clicked", clicked_handler)
    header_box.pack_end(button, False, False, 0)
    return header_box, button


def create_section_box(
    heading: str,
    *,
    margin_top: int = 12,
) -> tuple[Gtk.Box, Gtk.Box, Gtk.Label]:
    """Return a vertical section with a heading and content box."""

    section_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
    section_box.set_hexpand(True)

    heading_label = create_heading_label(heading, margin_top=max(0, margin_top))
    section_box.pack_start(heading_label, False, False, 0)

    content_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
    content_box.set_hexpand(True)
    content_box.set_margin_top(6)
    relation_set = content_box.get_accessible().ref_relation_set()
    relation_set.add(
        Atk.Relation.new([heading_label.get_accessible()], Atk.RelationType.LABELLED_BY)
    )
    section_box.pack_start(content_box, True, True, 0)
    return section_box, content_box, heading_label


def add_section_content(section_box: Gtk.Box, content_box: Gtk.Box, child: Gtk.Widget) -> None:
    """Add content to a section, preserving child expansion semantics."""

    expand = child.get_vexpand()
    content_box.pack_start(child, expand, True, 0)
    if child.get_hexpand():
        content_box.set_hexpand(True)
        section_box.set_hexpand(True)
    if expand:
        content_box.set_vexpand(True)
        section_box.set_vexpand(True)


def add_labelled_by(widget: Gtk.Widget, label: Gtk.Label) -> None:
    """Add an accessible labelled-by relation from widget to label."""

    relation_set = widget.get_accessible().ref_relation_set()
    relation_set.add(Atk.Relation.new([label.get_accessible()], Atk.RelationType.LABELLED_BY))


def create_dialog(
    title: str,
    *,
    default_size: tuple[int, int] | None = None,
    buttons: Sequence[tuple[str, int]] | None = None,
    default_response: int | None = None,
    modal: bool = True,
    resizable: bool = True,
    content_margin: int = 12,
) -> Gtk.Dialog:
    """Return an Orca-styled dialog."""

    sync_appearance()
    dialog = Gtk.Dialog(title=title, modal=modal)
    dialog.get_style_context().add_class("orca-dialog")
    dialog.set_titlebar(create_headerbar(title))
    dialog.set_resizable(resizable)
    if default_size is not None:
        dialog.set_default_size(*default_size)

    if buttons is None:
        from . import guilabels  # pylint: disable=import-outside-toplevel

        buttons = ((guilabels.BTN_CLOSE, Gtk.ResponseType.CLOSE),)
    for label, response in buttons:
        dialog.add_button(label, response)
    if default_response is not None:
        dialog.set_default_response(default_response)

    content_area = dialog.get_content_area()
    content_area.get_style_context().add_class("orca-dialog-content")
    content_area.set_border_width(content_margin)

    action_area = dialog.get_action_area()
    action_area.get_style_context().add_class("orca-dialog-content")
    action_area.get_style_context().add_class("orca-dialog-action-area")
    action_area.set_margin_start(content_margin)
    action_area.set_margin_end(content_margin)
    action_area.set_margin_top(content_margin)
    action_area.set_margin_bottom(content_margin)
    return dialog


def create_header_bar_dialog(
    title: str,
    cancel_label: str,
    ok_label: str,
    *,
    transient_for: Gtk.Widget | None = None,
    width: int = 600,
) -> tuple[Gtk.Dialog, Gtk.Button]:
    """Return a header-bar dialog with Cancel and suggested OK-style actions."""

    sync_appearance()
    parent = transient_for if isinstance(transient_for, Gtk.Window) else None
    dialog = Gtk.Dialog(
        title=title,
        transient_for=parent,
        modal=True,
        destroy_with_parent=True,
        use_header_bar=True,
    )
    dialog.set_default_size(width, -1)
    dialog.set_deletable(False)
    dialog.get_style_context().add_class("orca-dialog")

    header_bar = dialog.get_header_bar()
    header_bar.get_style_context().add_class("orca-panel-headerbar")

    cancel_button = Gtk.Button.new_with_mnemonic(cancel_label)
    cancel_button.connect("clicked", lambda _button: dialog.response(Gtk.ResponseType.CANCEL))
    header_bar.pack_start(cancel_button)

    ok_button = Gtk.Button.new_with_mnemonic(ok_label)
    ok_button.get_style_context().add_class("suggested-action")
    ok_button.connect("clicked", lambda _button: dialog.response(Gtk.ResponseType.OK))
    ok_button.set_can_default(True)
    header_bar.pack_end(ok_button)

    content_area = dialog.get_content_area()
    content_area.get_style_context().add_class("orca-dialog-content")
    content_area.set_border_width(24)
    content_area.set_spacing(18)

    return dialog, ok_button


def _separator_header_func(row: Gtk.ListBoxRow, before: Gtk.ListBoxRow | None, _data) -> None:
    """Add a separator before rows after the first one."""

    if before is not None:
        row.set_header(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))


def set_listbox_separators(listbox: Gtk.ListBox) -> None:
    """Add separators between rows in listbox."""

    listbox.set_header_func(_separator_header_func, None)


class FocusManagedListBox(Gtk.ListBox):
    """ListBox that manages focus for interactive widgets in rows."""

    def __init__(self, heading: str = ""):
        super().__init__()
        self.set_selection_mode(Gtk.SelectionMode.NONE)
        self.get_style_context().add_class("frame")
        self.set_can_focus(False)
        set_listbox_separators(self)

        self._widgets: list[Gtk.Widget] = []
        self._rows: list[Gtk.ListBoxRow] = []
        self._exiting_backward = [False]
        self._container: Gtk.Box | None = None

        if heading:
            label = create_heading_label(heading)
            self._container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            self._container.pack_start(label, False, False, 0)
            self._container.pack_start(self, False, False, 0)
            relation_set = self.get_accessible().ref_relation_set()
            relation_set.add(
                Atk.Relation.new([label.get_accessible()], Atk.RelationType.LABELLED_BY),
            )

    def get_container(self) -> Gtk.Widget:
        """Return the container to attach to a grid, or self if no heading."""

        return self._container or self

    def add_row_with_widget(self, row: Gtk.ListBoxRow, widget: Gtk.Widget) -> None:
        """Add a row with its associated interactive widget."""

        self.add_row_with_widgets(row, [widget])

    def add_row_with_widgets(self, row: Gtk.ListBoxRow, widgets: Sequence[Gtk.Widget]) -> None:
        """Add a row with associated interactive widgets."""

        if not widgets:
            self.add(row)
            self._rows.append(row)
            return

        for widget in widgets:
            widget.connect("key-press-event", self._on_widget_key_press)
        row.connect("focus-in-event", self._on_row_focus_in, widgets[0])

        self.add(row)
        self._rows.append(row)
        self._widgets.extend(widgets)

    def get_last_row(self) -> Gtk.ListBoxRow | None:
        """Return the last row that was added, or None if no rows."""

        if self._rows:
            return self._rows[-1]
        return None

    def _focus_next_sensitive_widget(self, widget: Gtk.Widget) -> bool:
        """Focus the next sensitive widget after the given one."""

        try:
            current_index = self._widgets.index(widget)
            for next_index in range(current_index + 1, len(self._widgets)):
                if self._widgets[next_index].get_sensitive():
                    self._widgets[next_index].grab_focus()
                    return True
        except ValueError:
            pass
        return False

    def _focus_prev_sensitive_widget(self, widget: Gtk.Widget) -> bool:
        """Focus the previous sensitive widget before the given one."""

        try:
            current_index = self._widgets.index(widget)
            for prev_index in range(current_index - 1, -1, -1):
                if self._widgets[prev_index].get_sensitive():
                    self._widgets[prev_index].grab_focus()
                    return True
            if self._rows:
                self._exiting_backward[0] = True
                self._rows[0].grab_focus()
        except ValueError:
            pass
        return False

    def _navigate_left_from_widget(self, _widget: Gtk.Widget) -> bool:
        """Handle Left arrow from widget. Subclasses can override."""

        return False

    def _on_widget_key_press(self, widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
        """Handle Tab, Shift+Tab, and Left arrow to navigate."""

        if event.keyval == Gdk.KEY_Tab:
            return self._focus_next_sensitive_widget(widget)
        if event.keyval == Gdk.KEY_Left:
            return self._navigate_left_from_widget(widget)
        if event.keyval == Gdk.KEY_ISO_Left_Tab:
            return self._focus_prev_sensitive_widget(widget)
        return False

    def _on_row_focus_in(self, _row, _event, widget: Gtk.Widget) -> bool:
        """Redirect focus from row to widget."""

        if self._exiting_backward[0]:
            self._exiting_backward[0] = False
            return False

        def activate_widget():
            widget.grab_focus()

        GLib.idle_add(activate_widget)
        return False


def create_framed_listbox(
    *,
    selection_mode: Gtk.SelectionMode = Gtk.SelectionMode.NONE,
    accessible_name: str = "",
    separators: bool = False,
) -> Gtk.ListBox:
    """Return an Orca-styled framed listbox."""

    listbox = Gtk.ListBox()
    listbox.set_selection_mode(selection_mode)
    listbox.get_style_context().add_class("frame")
    if accessible_name:
        listbox.get_accessible().set_name(accessible_name)
    if separators:
        set_listbox_separators(listbox)
    return listbox


def create_info_row(
    message: str,
    icon_name: str = "dialog-information-symbolic",
) -> Gtk.ListBoxRow:
    """Return an informational listbox row."""

    row = Gtk.ListBoxRow()
    row.set_activatable(False)
    row.set_can_focus(True)

    hbox = create_row_box()

    icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DND)
    icon.set_valign(Gtk.Align.START)
    hbox.pack_start(icon, False, False, 0)

    icon_accessible = icon.get_accessible()
    if icon_accessible:
        icon_accessible.set_role(Atk.Role.IMAGE)

    label = Gtk.Label(label=message, xalign=0)
    label.set_line_wrap(True)
    label.set_max_width_chars(60)
    label.set_hexpand(True)
    hbox.pack_start(label, True, True, 0)

    row.add(hbox)

    row_accessible = row.get_accessible()
    if row_accessible:
        row_accessible.set_name(message)
        row_accessible.set_role(Atk.Role.LABEL)

    return row


def create_info_listbox(
    message: str,
    icon_name: str = "dialog-information-symbolic",
) -> Gtk.ListBox:
    """Return a framed listbox containing one informational row."""

    listbox = create_framed_listbox()
    listbox_accessible = listbox.get_accessible()
    if listbox_accessible:
        listbox_accessible.set_role(Atk.Role.PANEL)
    listbox.add(create_info_row(message, icon_name))
    listbox.set_margin_bottom(12)
    return listbox


def create_key_value_row(label_text: str, value_text: str) -> Gtk.ListBoxRow:
    """Return a non-activatable row with a label and value."""

    row = Gtk.ListBoxRow()
    row.set_activatable(False)

    hbox = create_row_box()
    label = Gtk.Label(label=f"{label_text}: {value_text}", xalign=0)
    label.set_hexpand(True)
    label.set_line_wrap(True)
    label.set_max_width_chars(72)
    hbox.pack_start(label, True, True, 0)

    row.add(hbox)
    return row


def create_key_link_row(label_text: str, uri: str) -> Gtk.ListBoxRow:
    """Return a non-activatable row with a label and link button."""

    row = Gtk.ListBoxRow()
    row.set_activatable(False)

    hbox = create_row_box(spacing=0)
    label = Gtk.Label(label=f"{label_text}: ", xalign=0)
    link_button = Gtk.LinkButton.new_with_label(uri, uri)
    link_button.get_accessible().set_name(f"{label_text}: {uri}")
    link_button.set_relief(Gtk.ReliefStyle.NONE)

    def focus_link() -> bool:
        link_button.grab_focus()
        return False

    def on_row_focus_in(*_args) -> bool:
        GLib.idle_add(focus_link)
        return False

    row.connect("focus-in-event", on_row_focus_in)
    hbox.pack_start(label, False, False, 0)
    hbox.pack_start(link_button, False, False, 0)

    row.add(hbox)
    return row


def create_row_structure(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    include_top_separator: bool = True,
    label_text: str | None = None,
    widget: Gtk.Widget | None = None,
    label_xalign: float | None = None,
    label_halign: Gtk.Align | None = None,
    label_hexpand: bool = True,
    widget_expand: bool = False,
    label_size_group: Gtk.SizeGroup | None = None,
    set_mnemonic_widget: bool = True,
) -> tuple[Gtk.ListBoxRow, Gtk.Box, Gtk.Label | None]:
    """Return a standard listbox row structure."""

    row = Gtk.ListBoxRow()
    row.set_activatable(False)

    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
    if include_top_separator:
        vbox.pack_start(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL), False, False, 0)

    hbox = create_row_box()

    label = None
    if label_text and widget:
        label = Gtk.Label(label=label_text)
        label.set_use_underline(True)
        if label_xalign is not None:
            label.set_xalign(label_xalign)
        if label_halign is not None:
            label.set_halign(label_halign)
        label.set_hexpand(label_hexpand)
        if label_size_group is not None:
            label_size_group.add_widget(label)
        hbox.pack_start(label, label_hexpand, label_hexpand, 0)
        if set_mnemonic_widget:
            label.set_mnemonic_widget(widget)
        hbox.pack_end(widget, widget_expand, widget_expand, 0)

    vbox.pack_start(hbox, False, False, 0)
    row.add(vbox)
    return row, hbox, label


def create_switch_control(
    changed_handler: Callable[..., None] | None = None,
    state: bool | None = None,
    accessible_name: str = "",
    handler_args: tuple[object, ...] = (),
) -> Gtk.Switch:
    """Return an Orca-styled switch."""

    switch = Gtk.Switch()
    switch.set_valign(Gtk.Align.CENTER)
    if state is not None:
        switch.set_active(state)
    if accessible_name:
        switch.get_accessible().set_name(accessible_name)
    switch.get_accessible().set_role(Atk.Role.SWITCH)
    if changed_handler is not None:
        switch.connect("notify::active", changed_handler, *handler_args)
    return switch


def create_switch_row(
    label_text: str,
    changed_handler: Callable[..., None],
    state: bool,
    include_top_separator: bool = True,
    label_size_group: Gtk.SizeGroup | None = None,
) -> tuple[Gtk.ListBoxRow, Gtk.Switch, Gtk.Label]:
    """Return a standard listbox row with a label and switch."""

    switch = create_switch_control(changed_handler, state)
    row, _hbox, label = create_row_structure(
        include_top_separator,
        label_text,
        switch,
        label_xalign=0,
        label_size_group=label_size_group,
    )
    assert label is not None
    return row, switch, label


def create_spin_button_row(
    label_text: str,
    adjustment: Gtk.Adjustment,
    changed_handler: Callable[[Gtk.SpinButton], None] | None = None,
    include_top_separator: bool = True,
    digits: int = 0,
    label_size_group: Gtk.SizeGroup | None = None,
) -> tuple[Gtk.ListBoxRow, Gtk.SpinButton, Gtk.Label]:
    """Return a standard listbox row with a label and spin button."""

    spin = Gtk.SpinButton(adjustment=adjustment)
    spin.set_digits(digits)
    spin.set_valign(Gtk.Align.CENTER)
    if changed_handler is not None:
        spin.connect("value-changed", changed_handler)

    row, _hbox, label = create_row_structure(
        include_top_separator,
        label_text,
        spin,
        label_xalign=0,
        label_size_group=label_size_group,
    )
    assert label is not None
    return row, spin, label


def create_range_adjustment(
    value: float,
    minimum: float,
    maximum: float,
    *,
    steps: int = 100,
    page_steps: int = 10,
) -> Gtk.Adjustment:
    """Return an adjustment for numeric range controls."""

    extent = maximum - minimum
    step_increment = extent / steps if steps else extent
    page_increment = extent / page_steps if page_steps else extent
    return Gtk.Adjustment(
        value=value,
        lower=minimum,
        upper=maximum,
        step_increment=step_increment,
        page_increment=page_increment,
    )


def create_horizontal_size_group() -> Gtk.SizeGroup:
    """Return a horizontal size group."""

    return Gtk.SizeGroup(mode=Gtk.SizeGroupMode.HORIZONTAL)


def create_slider_row(
    label_text: str,
    adjustment: Gtk.Adjustment,
    changed_handler: Callable[[Gtk.Scale], None] | None = None,
    include_top_separator: bool = True,
    digits: int = 0,
    label_size_group: Gtk.SizeGroup | None = None,
) -> tuple[Gtk.ListBoxRow, Gtk.Scale, Gtk.Label]:
    """Return a standard listbox row with a label and horizontal slider."""

    scale = Gtk.Scale(orientation=Gtk.Orientation.HORIZONTAL, adjustment=adjustment)
    scale.set_digits(digits)
    scale.set_draw_value(True)
    scale.set_value_pos(Gtk.PositionType.RIGHT)
    scale.set_hexpand(True)
    scale.set_valign(Gtk.Align.CENTER)
    if changed_handler is not None:
        scale.connect("value-changed", changed_handler)

    row, _hbox, label = create_row_structure(
        include_top_separator,
        label_text,
        scale,
        label_halign=Gtk.Align.START,
        label_hexpand=False,
        widget_expand=True,
        label_size_group=label_size_group,
    )

    assert label is not None
    return row, scale, label


def create_range_slider_row(
    label_text: str,
    value: float,
    minimum: float,
    maximum: float,
    *,
    changed_handler: Callable[[Gtk.Scale], None] | None = None,
    include_top_separator: bool = True,
    digits: int = 0,
    steps: int = 100,
    page_steps: int = 10,
    label_size_group: Gtk.SizeGroup | None = None,
) -> tuple[Gtk.ListBoxRow, Gtk.Scale, Gtk.Label]:
    """Return a slider row for a numeric range."""

    adjustment = create_range_adjustment(
        value,
        minimum,
        maximum,
        steps=steps,
        page_steps=page_steps,
    )
    return create_slider_row(
        label_text,
        adjustment,
        changed_handler=changed_handler,
        include_top_separator=include_top_separator,
        digits=digits,
        label_size_group=label_size_group,
    )


def create_color_button_row(
    label_text: str,
    changed_handler: Callable[[Gtk.ColorButton], None] | None = None,
    include_top_separator: bool = True,
    use_alpha: bool = False,
    label_size_group: Gtk.SizeGroup | None = None,
) -> tuple[Gtk.ListBoxRow, Gtk.ColorButton, Gtk.Label]:
    """Return a standard listbox row with a label and color button."""

    color_button = Gtk.ColorButton()
    color_button.set_use_alpha(use_alpha)
    color_button.set_valign(Gtk.Align.CENTER)
    if changed_handler is not None:
        color_button.connect("color-set", changed_handler)

    row, _hbox, label = create_row_structure(
        include_top_separator,
        label_text,
        color_button,
        label_xalign=0,
        label_size_group=label_size_group,
    )
    assert label is not None
    return row, color_button, label


def set_color_button_hex(button: Gtk.ColorButton, hex_color: str) -> None:
    """Set a color button's color from a hex string."""

    button.set_color(Gdk.color_parse(hex_color))  # type: ignore[arg-type]


def rgba_to_hex(rgba: Gdk.RGBA) -> str:
    """Return a hex color string for rgba's red, green, and blue channels."""

    return f"#{int(rgba.red * 255):02x}{int(rgba.green * 255):02x}{int(rgba.blue * 255):02x}"


def create_entry(
    text: str = "",
    *,
    activates_default: bool = False,
    changed_handler: Callable[[Gtk.Entry], None] | None = None,
    activate_handler: Callable[[Gtk.Entry], None] | None = None,
    size_request: tuple[int, int] | None = None,
) -> Gtk.Entry:
    """Return a standard entry field."""

    entry = Gtk.Entry()
    if text:
        entry.set_text(text)
    entry.set_activates_default(activates_default)
    if size_request is not None:
        entry.set_size_request(*size_request)
    if changed_handler is not None:
        entry.connect("changed", changed_handler)
    if activate_handler is not None:
        entry.connect("activate", activate_handler)
    return entry


def create_labeled_entry_row(
    label_text: str,
    entry: Gtk.Entry,
    include_top_separator: bool = True,
    label_size_group: Gtk.SizeGroup | None = None,
) -> Gtk.ListBoxRow:
    """Return a standard listbox row with a label and entry field."""

    row, hbox, _label = create_row_structure(include_top_separator)
    hbox.pack_start(
        create_labeled_entry_box(label_text, entry, label_size_group),
        True,
        True,
        0,
    )
    return row


def create_labeled_entry_box(
    label_text: str,
    entry: Gtk.Entry,
    label_size_group: Gtk.SizeGroup | None = None,
    margin: int = 0,
) -> Gtk.Box:
    """Return a horizontal box with a label and entry field."""

    hbox = create_row_box(margin=margin)
    label = Gtk.Label(label=label_text, xalign=0)
    label.set_use_underline(True)
    hbox.pack_start(label, False, False, 0)

    if label_size_group:
        label_size_group.add_widget(label)

    entry.set_hexpand(True)
    label.set_mnemonic_widget(entry)
    hbox.pack_start(entry, True, True, 0)
    return hbox


def create_combo_box(
    model: Gtk.ListStore,
    changed_handler: Callable[[Gtk.ComboBox], None],
    *,
    valign: Gtk.Align | None = None,
) -> Gtk.ComboBox:
    """Return a combo box with Orca's standard text renderer setup."""

    combo = Gtk.ComboBox()
    combo.set_model(model)
    renderer = Gtk.CellRendererText()
    combo.pack_start(renderer, True)
    combo.add_attribute(renderer, "text", 0)
    combo.set_hexpand(False)
    if valign is not None:
        combo.set_valign(valign)
    combo.connect("changed", changed_handler)
    return combo


def create_combo_box_row(
    label_text: str,
    model: Gtk.ListStore,
    changed_handler: Callable[[Gtk.ComboBox], None],
    include_top_separator: bool = True,
    label_size_group: Gtk.SizeGroup | None = None,
) -> tuple[Gtk.ListBoxRow, Gtk.ComboBox, Gtk.Label]:
    """Return a standard listbox row with a label and combo box."""

    combo = create_combo_box(model, changed_handler, valign=Gtk.Align.CENTER)
    row, _hbox, label = create_row_structure(
        include_top_separator,
        label_text,
        combo,
        label_halign=Gtk.Align.START,
        label_size_group=label_size_group,
    )
    assert label is not None
    return row, combo, label


def enable_first_letter_nav(combo: Gtk.ComboBoxText) -> None:
    """Enable first-letter navigation on a ComboBoxText."""

    def on_key_press(_widget: Gtk.Widget, event: Any) -> bool:
        char = event.string.lower()
        if not char or not char.isalpha():
            return False
        model = combo.get_model()
        current = max(combo.get_active(), 0)
        for offset in range(1, len(model) + 1):
            idx = (current + offset) % len(model)
            text = (model[idx][0] or "").lower()
            if text.startswith(char):
                combo.set_active(idx)
                return True
        return False

    combo.connect("key-press-event", on_key_press)


def create_combo_box_text_row(
    label_text: str,
    items: Sequence[tuple[str, str]],
    include_top_separator: bool = True,
    changed_handler: Callable[[Gtk.ComboBoxText], None] | None = None,
    label_size_group: Gtk.SizeGroup | None = None,
) -> tuple[Gtk.ListBoxRow, Gtk.ComboBoxText, Gtk.Label | None]:
    """Return a listbox row with a label and ComboBoxText populated from items."""

    combo = Gtk.ComboBoxText()
    for item_id, display_text in items:
        combo.append(item_id, display_text)
    if items:
        combo.set_active(0)
    enable_first_letter_nav(combo)
    if changed_handler is not None:
        combo.connect("changed", changed_handler)

    row, _hbox, label = create_row_structure(
        include_top_separator,
        label_text,
        combo,
        label_xalign=0,
        label_halign=Gtk.Align.START,
        label_size_group=label_size_group,
    )

    return row, combo, label


def create_button_row(
    label_text: str,
    icon_name: str | None,
    clicked_handler: Callable[[Gtk.Button], None],
    include_top_separator: bool = True,
) -> tuple[Gtk.ListBoxRow, Gtk.Button, Gtk.Label]:
    """Return a standard listbox row with a label and button."""

    button = Gtk.Button()
    if icon_name:
        icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DND)
        button.set_image(icon)
    button.set_valign(Gtk.Align.CENTER)
    button.connect("clicked", clicked_handler)

    row, _hbox, label = create_row_structure(
        include_top_separator,
        label_text,
        button,
        label_halign=Gtk.Align.START,
        set_mnemonic_widget=False,
    )

    assert label is not None
    return row, button, label


def create_flat_icon_button(
    icon_name: str,
    accessible_name: str,
    icon_size: Gtk.IconSize = Gtk.IconSize.DND,
) -> Gtk.Button:
    """Return a flat icon button with an accessible name."""

    button = Gtk.Button.new_from_icon_name(icon_name, icon_size)
    button.set_relief(Gtk.ReliefStyle.NONE)
    button.get_accessible().set_name(accessible_name)
    return button


def set_stacked_label_text(
    label: Gtk.Label,
    primary_text: str,
    secondary_text: str = "",
    detail_text: str = "",
) -> None:
    """Set text for a stacked action-row label."""

    lines = [primary_text]
    lines.extend(text for text in (secondary_text, detail_text) if text)
    label.set_text("\n".join(lines))


def create_action_list_row(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    primary_text: str,
    secondary_text: str,
    actions: Sequence[ListRowAction] = (),
    include_top_separator: bool = True,
    primary_label_size_group: Gtk.SizeGroup | None = None,
    stack_labels: bool = False,
    detail_text: str = "",
) -> tuple[Gtk.ListBoxRow, Gtk.Label, Gtk.Label, dict[str, Gtk.Button]]:
    """Return a list row with labels and trailing action buttons."""

    row = Gtk.ListBoxRow()
    row.set_activatable(False)

    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
    if include_top_separator:
        vbox.pack_start(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL), False, False, 0)

    hbox = create_row_box()

    primary_label = Gtk.Label(label=primary_text, xalign=0)
    secondary_label = Gtk.Label(label=secondary_text, xalign=0)
    if stack_labels:
        primary_label.set_hexpand(True)
        primary_label.set_line_wrap(True)
        primary_label.set_max_width_chars(60)
        set_stacked_label_text(primary_label, primary_text, secondary_text, detail_text)
        hbox.pack_start(primary_label, True, True, 0)
    else:
        if primary_label_size_group:
            primary_label_size_group.add_widget(primary_label)
        hbox.pack_start(primary_label, False, False, 0)
        secondary_label.set_hexpand(True)
        hbox.pack_start(secondary_label, True, True, 0)

    action_buttons: dict[str, Gtk.Button] = {}
    button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
    for action in actions:
        if action.icon_name:
            button = Gtk.Button.new_from_icon_name(action.icon_name, Gtk.IconSize.DND)
            button.set_relief(Gtk.ReliefStyle.NONE)
            button.get_accessible().set_name(action.label)
        else:
            button = Gtk.Button(label=action.label)
        button.connect("clicked", action.callback)
        button_box.pack_start(button, False, False, 0)
        action_buttons[action.name] = button

    hbox.pack_end(button_box, False, False, 0)
    vbox.pack_start(hbox, False, False, 0)
    row.add(vbox)
    return row, primary_label, secondary_label, action_buttons


def create_scrolled_window(
    child: Gtk.Widget | None = None,
    *,
    hscroll_policy: Gtk.PolicyType | None = None,
    vscroll_policy: Gtk.PolicyType | None = None,
    size_request: tuple[int, int] | None = None,
    hexpand: bool | None = None,
    vexpand: bool | None = None,
    framed: bool = True,
) -> Gtk.ScrolledWindow:
    """Return a scrolled window for Orca dialogs."""

    scrolled_window = Gtk.ScrolledWindow()
    if framed:
        scrolled_window.get_style_context().add_class("frame")
        scrolled_window.set_shadow_type(Gtk.ShadowType.IN)
    if hscroll_policy is not None or vscroll_policy is not None:
        scrolled_window.set_policy(
            hscroll_policy or Gtk.PolicyType.AUTOMATIC,
            vscroll_policy or Gtk.PolicyType.AUTOMATIC,
        )
    if size_request is not None:
        scrolled_window.set_size_request(*size_request)
    if hexpand is not None:
        scrolled_window.set_hexpand(hexpand)
    if vexpand is not None:
        scrolled_window.set_vexpand(vexpand)
    if child is not None:
        scrolled_window.add(child)  # pylint: disable=no-member
    return scrolled_window


def create_preferences_scrolled_window(child: Gtk.Widget | None = None) -> Gtk.ScrolledWindow:
    """Return a scrolled window for preferences-grid content."""

    return create_scrolled_window(
        child or Gtk.Box(),
        hscroll_policy=Gtk.PolicyType.NEVER,
        vscroll_policy=Gtk.PolicyType.AUTOMATIC,
        size_request=(700, -1),
        hexpand=True,
        vexpand=True,
    )


def create_tree_view() -> tuple[Gtk.ScrolledWindow, Gtk.TreeView]:
    """Return a framed scrolled window containing a tree view."""

    tree = Gtk.TreeView()
    tree.set_hexpand(True)
    tree.set_vexpand(True)
    return create_scrolled_window(tree), tree


def create_tree_view_dialog(
    title: str,
    *,
    column_headers: Sequence[str] | None = None,
    text_column_offset: int = 0,
    sort_empty_headers: bool = False,
    default_size: tuple[int, int] | None = None,
    buttons: Sequence[tuple[str, int]] | None = None,
    default_response: int | None = None,
    modal: bool = True,
    resizable: bool = True,
    content_margin: int = 12,
) -> tuple[Gtk.Dialog, Gtk.TreeView]:
    """Return an Orca-styled dialog containing a framed tree view."""

    dialog = create_dialog(
        title,
        default_size=default_size,
        buttons=buttons,
        default_response=default_response,
        modal=modal,
        resizable=resizable,
        content_margin=content_margin,
    )

    grid = Gtk.Grid()
    dialog.get_content_area().add(grid)

    scrolled_window, tree = create_tree_view()
    grid.add(scrolled_window)  # pylint: disable=no-member
    add_tree_view_columns(
        tree,
        column_headers or (),
        text_column_offset=text_column_offset,
        sort_empty_headers=sort_empty_headers,
    )
    return dialog, tree


def add_tree_view_columns(
    tree: Gtk.TreeView,
    column_headers: Sequence[str],
    *,
    text_column_offset: int = 0,
    sort_empty_headers: bool = False,
) -> None:
    """Add sortable text columns to a tree view."""

    for i, header in enumerate(column_headers):
        cell = Gtk.CellRendererText()
        model_column = i + text_column_offset
        column = Gtk.TreeViewColumn(header, cell, text=model_column)
        tree.append_column(column)
        if header or sort_empty_headers:
            column.set_sort_column_id(model_column)


def apply_appearance(
    interface_settings: Gio.Settings,
    a11y_settings: Gio.Settings,
    gtk_settings: Gtk.Settings,
    screen: Gdk.Screen,
    providers: AppearanceProviders,
) -> None:
    """Apply color-scheme, high-contrast, and status-shapes settings."""

    prefer_dark = interface_settings.get_string("color-scheme") == "prefer-dark"
    gtk_settings.set_property("gtk-application-prefer-dark-theme", prefer_dark)

    theme = gtk_settings.get_property("gtk-theme-name")
    if prefer_dark and theme == "HighContrast":
        gtk_settings.set_property("gtk-theme-name", "HighContrastInverse")
    elif not prefer_dark and theme == "HighContrastInverse":
        gtk_settings.set_property("gtk-theme-name", "HighContrast")

    priority = _style_provider_priority()
    if priority is None:
        return
    priority += 1
    conditional_providers: list[tuple[Gtk.CssProvider, bool]] = [
        (providers.hc, a11y_settings.get_boolean("high-contrast")),
        (providers.dark, prefer_dark),
    ]
    if providers.shapes is not None:
        conditional_providers.append(
            (providers.shapes, a11y_settings.get_boolean("show-status-shapes")),
        )
    for provider, enabled in conditional_providers:
        if enabled:
            Gtk.StyleContext.add_provider_for_screen(screen, provider, priority)
        else:
            Gtk.StyleContext.remove_provider_for_screen(screen, provider)



# Orca
#
# Copyright 2024-2026 Igalia, S.L.
# Copyright 2024 GNOME Foundation Inc.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Wrapper for the Atspi.Component interface."""

from __future__ import annotations

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

from . import debug
from .ax_object import AXObject


class AXComponent:
    """Wrapper for the Atspi.Component interface."""

    @staticmethod
    def get_position(obj: Atspi.Accessible) -> tuple[int, int]:
        """Returns the x, y position tuple of obj with respect to its window."""

        if not AXObject.supports_component(obj):
            return -1, -1

        try:
            point = Atspi.Component.get_position(obj, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            msg = f"AXComponent: Exception in get_position: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return -1, -1

        if point is None:
            tokens = ["AXComponent: get_position failed for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return -1, -1

        return point.x, point.y

    @staticmethod
    def get_rect(obj: Atspi.Accessible) -> Atspi.Rect:
        """Returns the Atspi rect of obj with respect to its window."""

        if not AXObject.supports_component(obj):
            return Atspi.Rect()

        try:
            rect = Atspi.Component.get_extents(obj, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            msg = f"AXComponent: Exception in get_rect: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return Atspi.Rect()

        return rect

    @staticmethod
    def get_size(obj: Atspi.Accessible) -> tuple[int, int]:
        """Returns the width, height tuple of obj with respect to its window."""

        if not AXObject.supports_component(obj):
            return -1, -1

        try:
            point = Atspi.Component.get_size(obj, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            msg = f"AXComponent: Exception in get_position: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return -1, -1

        if point is None:
            tokens = ["AXComponent: get_size failed for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return -1, -1

        # An Atspi.Point object stores width in x and height in y.
        return point.x, point.y

    @staticmethod
    def object_contains_point(obj: Atspi.Accessible, x: int, y: int) -> bool:
        """Returns True if obj's rect contains the specified point."""

        if not AXObject.supports_component(obj):
            return False

        if AXObject.is_bogus(obj):
            return False

        try:
            result = Atspi.Component.contains(obj, x, y, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            msg = f"AXComponent: Exception in object_contains_point: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        tokens = ["AXComponent: ", obj, f"contains point {x}, {y}: {result}"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_object_at_point(obj: Atspi.Accessible, x: int, y: int) -> Atspi.Accessible | None:
        """Returns the child (or descendant?) of obj at the specified point."""

        if not AXObject.supports_component(obj):
            return None

        try:
            result = Atspi.Component.get_accessible_at_point(obj, x, y, Atspi.CoordType.WINDOW)
        except GLib.GError as error:
            msg = f"AXComponent: Exception in get_child_at_point: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return None

        tokens = ["AXComponent: Child of", obj, f"at {x}, {y} is", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def scroll_object_to_point(obj: Atspi.Accessible, x: int, y: int) -> bool:
        """Attempts to scroll obj to the specified point."""

        if not AXObject.supports_component(obj):
            return False

        try:
            result = Atspi.Component.scroll_to_point(obj, Atspi.CoordType.WINDOW, x, y)
        except GLib.GError as error:
            msg = f"AXComponent: Exception in scroll_object_to_point: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        tokens = ["AXComponent: Scrolled", obj, f"to {x}, {y}:", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def scroll_object_to_location(obj: Atspi.Accessible, location: Atspi.ScrollType) -> bool:
        """Attempts to scroll obj to the specified Atspi.ScrollType location."""

        if not AXObject.supports_component(obj):
            return False

        try:
            result = Atspi.Component.scroll_to(obj, location)
        except GLib.GError as error:
            msg = f"AXComponent: Exception in scroll_object_to_location: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        tokens = ["AXComponent: Scrolled", obj, "to", location, f": {result}"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result



# Orca
#
# Copyright 2004-2008 Sun Microsystems Inc.
# Copyright 2010 Joanmarie Diggs
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.


"""Custom script for pidgin."""

from __future__ import annotations

from typing import TYPE_CHECKING

from orca import chat_presenter, debug, messages, presentation_manager
from orca.ax_object import AXObject
from orca.ax_utilities import AXUtilities
from orca.scripts.toolkits import gtk

from .script_utilities import Utilities
from .speech_generator import SpeechGenerator

if TYPE_CHECKING:
    import gi

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi


class Script(gtk.Script):
    """Custom script for pidgin."""

    # Override the base class type annotation
    utilities: Utilities

    def _create_speech_generator(self) -> SpeechGenerator:
        """Creates and returns the speech generator for this script."""

        return SpeechGenerator(self)

    def get_utilities(self) -> Utilities:
        """Returns the utilities for this script."""

        return Utilities(self)

    def _on_children_added(self, event: Atspi.Event) -> bool:
        """Callback for object:children-changed:add accessibility events."""

        super()._on_children_added(event)
        if not AXUtilities.is_page_tab_list(event.source):
            return True

        AXObject.clear_cache(event.source, True, "to ensure tab info is current.")

        if AXUtilities.is_selected(event.any_data):
            msg = "PIDGIN: Not presenting addition of already-selected tab"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        # In the chat window, the frame name changes to reflect the active chat.
        # So if we don't have a matching tab, this isn't the chat window.
        frame = AXUtilities.find_ancestor(event.source, AXUtilities.is_frame)
        frame_name = AXObject.get_name(frame)
        for child in AXObject.iter_children(event.source):
            if frame_name == AXObject.get_name(child):
                break
        else:
            tokens = ["PIDGIN:", frame, "does not seem to be a chat window"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        line = messages.CHAT_NEW_TAB % AXObject.get_name(event.any_data)
        presentation_manager.get_manager().speak_accessible_text(event.any_data, line)
        return True

    def _on_expanded_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:expanded accessibility events."""

        # Overridden here because the event.source is in a hidden column.
        obj = event.source
        if self.chat.is_in_buddy_list(obj):
            obj = AXObject.get_next_sibling(obj)
            self.present_object(obj, prior_obj=obj)
            return True

        return super()._on_expanded_changed(event)

    def _on_text_inserted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:insert accessibility events."""

        if chat_presenter.get_presenter().present_inserted_text(self, event):
            return True

        return super()._on_text_inserted(event)



# Orca
#
# Copyright 2004-2009 Sun Microsystems Inc.
# Copyright 2010-2011 The Orca Team
# Copyright 2012 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-locals

"""The main module for the Orca screen reader."""

import os
import signal
import sys

import gi
from dasbus.connection import SessionMessageBus

gi.require_version("Atspi", "2.0")
gi.require_version("Gdk", "3.0")
from gi.repository import (
    Atspi,
    Gdk,  # pylint: disable=no-name-in-module
    GLib,
)

from . import (
    ax_device_manager,
    clipboard,
    command_manager,
    dbus_service,
    debug,
    debugging_tools_manager,
    event_manager,
    extension_loader,
    focus_manager,
    messages,
    mouse_presenter,
    orca_modifier_manager,
    presentation_manager,
    script_manager,
    systemd,
)
from .ax_utilities import AXUtilities


def load_user_settings(script=None, skip_reload_message=False, is_reload=True):
    """(Re)Loads the user settings module, re-initializing things such as speech if necessary."""

    debug.print_message(debug.LEVEL_INFO, "ORCA: Loading User Settings", True)

    if is_reload:
        presentation_manager.get_manager().shutdown_presenters()
        mouse_presenter.get_presenter().deactivate()

    event_manager.get_manager().pause_queuing(True, True, "Loading user settings.")

    if script is None:
        script = script_manager.get_manager().get_default_script()

    command_manager.get_manager().load_keyboard_layout()

    presentation_manager.get_manager().start_presenters()
    if is_reload and not skip_reload_message:
        presentation_manager.get_manager().speak_message(messages.SETTINGS_RELOADED)

    # TODO - JD: This ultimately belongs in an extension manager.
    presenter = mouse_presenter.get_presenter()
    if presenter.get_is_enabled():
        presenter.activate()

    # Handle the case where a change was made in the Orca Preferences dialog.
    orca_modifier_manager.get_manager().refresh_orca_modifiers("Loading user settings.")
    event_manager.get_manager().pause_queuing(False, False, "User settings loaded.")
    debug.print_message(debug.LEVEL_INFO, "ORCA: User Settings Loaded", True)

    manager = debugging_tools_manager.get_manager()
    manager.print_session_details()
    manager.print_running_applications(force=False)


def shutdown(_event=None, _signum=None):
    """Exits Orca. Returns True if shutdown ran to completion."""

    if hasattr(shutdown, "in_progress"):
        msg = "ORCA: Shutdown already in progress"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return False

    shutdown.in_progress = True

    def _timeout(_signum=None, _frame=None):
        msg = "TIMEOUT: something has hung. Aborting."
        debug.print_message(debug.LEVEL_SEVERE, msg, True)
        os.kill(os.getpid(), signal.SIGKILL)

    debug.print_message(debug.LEVEL_INFO, "ORCA: Shutting down", True)
    signal.signal(signal.SIGALRM, _timeout)
    signal.alarm(5)

    manager = presentation_manager.get_manager()
    manager.interrupt_presentation()
    manager.present_message(messages.STOP_ORCA)

    extension_loader.get_loader().shutdown_user_extensions()
    dbus_service.get_remote_controller().shutdown()
    orca_modifier_manager.get_manager().unset_orca_modifiers("Shutting down.")

    # Pause event queuing first so that it clears its queue and will not accept new
    # events. Then let the script manager unregister script event listeners as well
    # as key grabs. Finally deactivate the event manager and the device
    # manager, which will cause the Atspi.Device to be set to None.
    event_manager.get_manager().pause_queuing(True, True, "Shutting down.")
    script_manager.get_manager().deactivate()
    event_manager.get_manager().deactivate()

    # TODO - JD: This ultimately belongs in an extension manager.
    clipboard.get_presenter().deactivate()
    mouse_presenter.get_presenter().deactivate()

    presentation_manager.get_manager().shutdown_presenters()

    # pylint: disable-next=import-outside-toplevel
    from . import preferences_window

    preferences_window.close_preferences_gui_for_shutdown()

    ax_device_manager.get_manager().deactivate()
    systemd.get_manager().notify_stopping()
    signal.alarm(0)
    debug.print_message(debug.LEVEL_INFO, "ORCA: Quitting Atspi main event loop", True)
    Atspi.event_quit()  # pylint: disable=no-value-for-parameter
    debug.print_message(debug.LEVEL_INFO, "ORCA: Shutdown complete", True)
    debug.shutdown()
    return True


def _setup_signal_handlers():
    """Sets up signal handlers for reload, shutdown, and preferences."""

    def _glib_shutdown_handler():
        msg = "ORCA: GLib handler received shutdown signal"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        shutdown()
        return GLib.SOURCE_REMOVE

    def _glib_reload_handler():
        msg = "ORCA: GLib handler received reload signal"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        systemd.get_manager().notify_reloading()
        load_user_settings()
        systemd.get_manager().notify_ready()
        return GLib.SOURCE_CONTINUE

    def _glib_show_preferences_handler():
        msg = "ORCA: GLib handler received show-preferences signal"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        from . import screen_reader_manager  # pylint: disable=import-outside-toplevel

        screen_reader_manager.get_manager().show_preferences_gui()
        return GLib.SOURCE_CONTINUE

    for signum, handler in (
        (signal.SIGINT, _glib_shutdown_handler),
        (signal.SIGTERM, _glib_shutdown_handler),
        (signal.SIGHUP, _glib_reload_handler),
        (signal.SIGUSR1, _glib_show_preferences_handler),
    ):
        GLib.unix_signal_add(GLib.PRIORITY_HIGH, signum, handler)

    # We have historically shut down gracefully on SIGQUIT and want to continue doing so.
    # But GLib.unix_signal_add only supports SIGHUP/SIGINT/SIGTERM/SIGUSR1/SIGUSR2/SIGWINCH,
    # so SIGQUIT uses a raw handler that defers the shutdown to the main loop.
    def _quit_on_signal(_signum, _frame):
        GLib.idle_add(_glib_shutdown_handler, priority=GLib.PRIORITY_HIGH)

    signal.signal(signal.SIGQUIT, _quit_on_signal)


def _ensure_accessibility_enabled() -> int | None:
    """Ensures the accessibility bus is enabled. Returns error code or None on success."""

    try:
        bus = SessionMessageBus()
        proxy = bus.get_proxy("org.a11y.Bus", "/org/a11y/bus", "org.freedesktop.DBus.Properties")
        enabled = proxy.Get("org.a11y.Status", "IsEnabled")
        msg = f"ORCA: Accessibility enabled: {enabled}"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        if not enabled:
            msg = "ORCA: Enabling accessibility."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            proxy.Set("org.a11y.Status", "IsEnabled", GLib.Variant("b", True))
    except GLib.GError as error:
        msg = f"ORCA: Could not connect to D-Bus session bus: {error}"
        debug.print_message(debug.LEVEL_SEVERE, msg, True)
        print(msg, file=sys.stderr)  # noqa: T201
        return 1
    return None


def _activate_services():
    """Activates Orca services and attempts to find the focused object."""

    ax_device_manager.get_manager().activate()
    event_manager.get_manager().activate()
    script_manager.get_manager().activate()
    window = AXUtilities.find_active_window()
    if window and not focus_manager.get_manager().get_locus_of_focus():
        app = AXUtilities.get_application(window)
        focus_manager.get_manager().set_active_window(
            window,
            app,
            set_window_as_focus=True,
            notify_script=True,
        )

        focused_object = focus_manager.get_manager().find_focused_object()
        if focused_object:
            focus_manager.get_manager().set_locus_of_focus(None, focused_object)

        script = script_manager.get_manager().get_script(app, focused_object or window)
        script_manager.get_manager().set_active_script(script, "Launching.")

    clipboard.get_presenter().activate()
    Gdk.notify_startup_complete()  # pylint: disable=no-value-for-parameter
    systemd.get_manager().notify_ready()
    extension_loader.get_loader().notify_user_extensions_ready()


def main():
    """The main entry point for Orca."""

    _setup_signal_handlers()
    systemd.get_manager().start_watchdog()

    error = _ensure_accessibility_enabled()
    if error is not None:
        return error

    ax_device_manager.get_manager().activate()
    load_user_settings(is_reload=False)

    is_systemd_managed = systemd.get_manager().is_systemd_managed()
    msg = f"ORCA: Running under systemd: {is_systemd_managed}"
    debug.print_message(debug.LEVEL_INFO, msg, True)

    dbus_service.get_remote_controller().start()
    presentation_manager.get_manager().present_message(messages.START_ORCA)
    _activate_services()

    try:
        debug.print_message(debug.LEVEL_INFO, "ORCA: Starting Atspi main event loop", True)
        Atspi.event_main()  # pylint: disable=no-value-for-parameter
    except GLib.Error as error:
        msg = f"ORCA: Exception starting ATSPI registry: {error}"
        debug.print_message(debug.LEVEL_SEVERE, msg, True)
        os.kill(os.getpid(), signal.SIGKILL)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Orca
#
# Copyright 2013-2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Command definitions for caret navigation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import cmdnames, keybindings
from .command import Command, KeyboardCommand

if TYPE_CHECKING:
    from .caret_navigator import CaretNavigator


def get_commands(owner: CaretNavigator) -> list[Command]:
    """Returns commands for caret navigation."""

    kb_orca_f12 = keybindings.KeyBinding("F12", keybindings.ORCA_MODIFIER_MASK)
    kb_unmodified_right = keybindings.KeyBinding("Right", keybindings.NO_MODIFIER_MASK)
    kb_unmodified_left = keybindings.KeyBinding("Left", keybindings.NO_MODIFIER_MASK)
    kb_ctrl_right = keybindings.KeyBinding("Right", keybindings.CTRL_MODIFIER_MASK)
    kb_ctrl_left = keybindings.KeyBinding("Left", keybindings.CTRL_MODIFIER_MASK)
    kb_unmodified_down = keybindings.KeyBinding("Down", keybindings.NO_MODIFIER_MASK)
    kb_unmodified_up = keybindings.KeyBinding("Up", keybindings.NO_MODIFIER_MASK)
    kb_unmodified_end = keybindings.KeyBinding("End", keybindings.NO_MODIFIER_MASK)
    kb_unmodified_home = keybindings.KeyBinding("Home", keybindings.NO_MODIFIER_MASK)
    kb_ctrl_end = keybindings.KeyBinding("End", keybindings.CTRL_MODIFIER_MASK)
    kb_ctrl_home = keybindings.KeyBinding("Home", keybindings.CTRL_MODIFIER_MASK)

    enabled = owner.get_is_enabled() and not owner._suspended  # pylint: disable=protected-access
    toggle_enabled = not owner._suspended  # pylint: disable=protected-access

    return [
        KeyboardCommand(
            "toggle_enabled",
            owner.toggle_enabled,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_TOGGLE,
            desktop_keybinding=kb_orca_f12,
            laptop_keybinding=kb_orca_f12,
            enabled=toggle_enabled,
            is_group_toggle=True,
        ),
        KeyboardCommand(
            "next_character",
            owner.next_character,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_NEXT_CHAR,
            desktop_keybinding=kb_unmodified_right,
            laptop_keybinding=kb_unmodified_right,
            enabled=enabled,
        ),
        KeyboardCommand(
            "previous_character",
            owner.previous_character,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_PREV_CHAR,
            desktop_keybinding=kb_unmodified_left,
            laptop_keybinding=kb_unmodified_left,
            enabled=enabled,
        ),
        KeyboardCommand(
            "next_word",
            owner.next_word,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_NEXT_WORD,
            desktop_keybinding=kb_ctrl_right,
            laptop_keybinding=kb_ctrl_right,
            enabled=enabled,
        ),
        KeyboardCommand(
            "previous_word",
            owner.previous_word,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_PREV_WORD,
            desktop_keybinding=kb_ctrl_left,
            laptop_keybinding=kb_ctrl_left,
            enabled=enabled,
        ),
        KeyboardCommand(
            "next_line",
            owner.next_line,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_NEXT_LINE,
            desktop_keybinding=kb_unmodified_down,
            laptop_keybinding=kb_unmodified_down,
            enabled=enabled,
        ),
        KeyboardCommand(
            "previous_line",
            owner.previous_line,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_PREV_LINE,
            desktop_keybinding=kb_unmodified_up,
            laptop_keybinding=kb_unmodified_up,
            enabled=enabled,
        ),
        KeyboardCommand(
            "start_of_file",
            owner.start_of_file,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_FILE_START,
            desktop_keybinding=kb_ctrl_home,
            laptop_keybinding=kb_ctrl_home,
            enabled=enabled,
        ),
        KeyboardCommand(
            "end_of_file",
            owner.end_of_file,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_FILE_END,
            desktop_keybinding=kb_ctrl_end,
            laptop_keybinding=kb_ctrl_end,
            enabled=enabled,
        ),
        KeyboardCommand(
            "start_of_line",
            owner.start_of_line,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_LINE_START,
            desktop_keybinding=kb_unmodified_home,
            laptop_keybinding=kb_unmodified_home,
            enabled=enabled,
        ),
        KeyboardCommand(
            "end_of_line",
            owner.end_of_line,
            owner.GROUP_LABEL,
            cmdnames.CARET_NAVIGATION_LINE_END,
            desktop_keybinding=kb_unmodified_end,
            laptop_keybinding=kb_unmodified_end,
            enabled=enabled,
        ),
        KeyboardCommand(
            "toggle_layout_mode",
            owner.toggle_layout_mode,
            owner.GROUP_LABEL,
            cmdnames.TOGGLE_LAYOUT_MODE,
            enabled=enabled,
        ),
    ]



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2010 Orca Team.
# Copyright 2014-2025 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-boolean-expressions
# pylint: disable=too-many-branches
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-lines
# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-statements

"""Provides support for accessing user-agent-agnostic web-content."""

from __future__ import annotations

from typing import TYPE_CHECKING

from orca import (
    braille_presenter,
    caret_navigator,
    debug,
    document_presenter,
    focus_manager,
    input_event_manager,
    label_inference,
    live_region_presenter,
    messages,
    presentation_manager,
    say_all_presenter,
    speech_manager,
    speech_presenter,
    structural_navigator,
    table_navigator,
)
from orca.ax_event_synthesizer import AXEventSynthesizer
from orca.ax_object import AXObject
from orca.ax_text import AXText
from orca.ax_utilities import AXUtilities
from orca.ax_utilities_event import TextEventReason
from orca.ax_utilities_text import CaretSetReason, TextUnit
from orca.scripts import default
from orca.structural_navigator import NavigationMode

from .braille_generator import BrailleGenerator
from .script_utilities import Utilities
from .speech_generator import SpeechGenerator

if TYPE_CHECKING:
    from gi.repository import Atspi

    from orca.generator import PresentationReason


class Script(default.Script):
    """Provides support for accessing user-agent-agnostic web-content."""

    # Type annotations to override the base class types
    utilities: Utilities
    caret_navigator: caret_navigator.CaretNavigator

    def __init__(self, app: Atspi.Accessible) -> None:
        super().__init__(app)

        self.present_if_inactive: bool = False
        self._default_sn_mode = NavigationMode.DOCUMENT
        self._default_caret_navigation_enabled: bool = True

        self._loading_content = False
        self._last_mouse_button_context = None, -1

    def deactivate(self) -> None:
        """Called when this script is deactivated."""

        self._loading_content = False
        document_presenter.get_presenter().reset_find_announcement_state()
        self._last_mouse_button_context = None, -1
        self.utilities.clear_cached_objects()
        super().deactivate()

    def _create_braille_generator(self) -> BrailleGenerator:
        """Creates and returns the braille generator for this script."""

        return BrailleGenerator(self)

    def get_label_inference(self) -> label_inference.LabelInference | None:
        """Returns the label inference functionality for this script."""

        return label_inference.LabelInference(self)

    def _create_speech_generator(self) -> SpeechGenerator:
        """Creates and returns the speech generator for this script."""

        return SpeechGenerator(self)

    def get_utilities(self) -> Utilities:
        """Returns the utilities for this script."""

        return Utilities(self)

    def is_loading_content(self) -> bool:
        """Returns True if we're currently loading content."""

        return self._loading_content

    def say_character(self, obj: Atspi.Accessible, offset: int) -> None:
        """Speaks the character at the specified offset."""

        tokens = ["WEB: Say character for", obj, "offset", offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if not self.utilities.in_document_content(obj):
            msg = "WEB: Object is not in document content."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            super().say_character(obj, offset)
            return

        contents: list[tuple[Atspi.Accessible, int, int, str]] | None = None
        if self.utilities.treat_as_end_of_line(obj, offset) and AXObject.supports_text(obj):
            char = AXText.get_character_at_offset(obj, offset, ensure_whole_characters=True)[0]
            if char == "\ufffc":
                char = ""
            contents = [(obj, offset, offset + 1, char)]
        else:
            contents = self.utilities.get_character_contents_at_offset(obj, offset)

        if not contents:
            return

        speech_pres = speech_presenter.get_presenter()
        obj, start, _end, string = contents[0]
        if start > 0 and string == "\n":
            if speech_pres.get_speak_blank_lines():
                presentation_manager.get_manager().speak_message(messages.BLANK)
                return

        presenter = presentation_manager.get_manager()
        if string:
            speech_pres.present_text_attribute_state(obj, start)
            language, dialect = self.utilities.get_language_and_dialect_for_substring(
                obj, start, start + 1
            )
            presenter.speak_character(string, obj=obj, language=language, dialect=dialect)
        else:
            presenter.speak_contents(contents)

        AXUtilities.set_last_text_unit_spoken(TextUnit.CHAR)

    def say_word(self, obj: Atspi.Accessible, offset: int) -> None:
        """Speaks the word at the specified offset."""

        tokens = ["WEB: Say word for", obj, "offset", offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if not self.utilities.in_document_content(obj):
            msg = "WEB: Object is not in document content."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            super().say_word(obj, offset)
            return

        if input_event_manager.get_manager().last_event_was_right():
            offset -= 1

        word_contents = self.utilities.get_word_contents_at_offset(obj, offset, use_cache=True)
        text_obj, start_offset, _end_offset, _word = word_contents[0]

        speech_pres = speech_presenter.get_presenter()
        speech_pres.present_text_attribute_state(text_obj, start_offset)
        speech_pres.speak_word(self, obj, offset)
        AXUtilities.set_last_text_unit_spoken(TextUnit.WORD)

    def present_object(
        self,
        obj: Atspi.Accessible,
        offset: int | None = None,
        prior_obj: Atspi.Accessible | None = None,
        generate_speech: bool = True,
        generate_braille: bool = True,
        reason: PresentationReason | None = None,
    ) -> None:
        if obj is None:
            return

        if not self.utilities.in_document_content(obj) or AXUtilities.is_document(obj):
            super().present_object(
                obj,
                offset=offset,
                prior_obj=prior_obj,
                generate_speech=generate_speech,
                generate_braille=generate_braille,
                reason=reason,
            )
            return

        mode, _obj = focus_manager.get_manager().get_active_mode_and_object_of_interest()
        if mode in [focus_manager.OBJECT_NAVIGATOR, focus_manager.MOUSE_REVIEW]:
            super().present_object(
                obj,
                offset=offset,
                prior_obj=prior_obj,
                generate_speech=generate_speech,
                generate_braille=generate_braille,
                reason=reason,
            )
            return

        if AXUtilities.is_status_bar(obj) or AXUtilities.is_alert(obj):
            if not document_presenter.get_presenter().in_focus_mode(self.app):
                self.utilities.set_caret_position(obj, 0, reason=CaretSetReason.OBJECT_PRESENTATION)
            super().present_object(
                obj,
                offset=offset,
                prior_obj=prior_obj,
                generate_speech=generate_speech,
                generate_braille=generate_braille,
                reason=reason,
            )
            return

        prior_reason = None
        if caret_navigator.get_navigator().last_input_event_was_navigation_command():
            prior_reason = "caret navigation command"
        elif structural_navigator.get_navigator().last_input_event_was_navigation_command():
            prior_reason = "structural navigation command"
        elif table_navigator.get_navigator().last_input_event_was_navigation_command():
            prior_reason = "table navigation command"
        elif AXUtilities.get_table(obj):
            prior_reason = "object in table"

        if prior_reason:
            tokens = ["WEB: Using prior context for presentation:", prior_reason]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            prior_context = self.utilities.get_prior_context()
            if prior_context is not None:
                prior_obj, _prior_offset = prior_context

        document = self.utilities.get_document_for_object(obj)
        obj = AXEventSynthesizer.scroll_to_center(obj, start_offset=0, root=document)
        if obj is None:
            msg = "WEB: Scroll target was destroyed and could not be relocated. Skipping."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return

        # Editors like VSCode use the entry role for the code editor.
        if AXUtilities.is_entry(obj):
            if not document_presenter.get_presenter().in_focus_mode(self.app):
                self.utilities.set_caret_position(obj, 0, reason=CaretSetReason.OBJECT_PRESENTATION)
            super().present_object(
                obj,
                offset=offset,
                prior_obj=prior_obj,
                generate_speech=generate_speech,
                generate_braille=generate_braille,
                reason=reason,
            )
            return

        tokens = ["WEB: Presenting object", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # We shouldn't use cache in this method, because if the last thing we presented
        # included this object and offset (e.g. a Say All or Mouse Review), we're in
        # danger of presented irrelevant context.
        effective_offset = offset if offset is not None else 0
        contents = self.utilities.get_object_contents_at_offset(
            obj, effective_offset, use_cache=False
        )
        if (
            contents
            and contents[0]
            and not document_presenter.get_presenter().in_focus_mode(self.app)
        ):
            self.utilities.set_caret_position(
                contents[0][0], contents[0][1], reason=CaretSetReason.OBJECT_PRESENTATION
            )
        presenter = presentation_manager.get_manager()
        if generate_braille:
            presenter.display_contents(contents)
        if generate_speech:
            presenter.speak_contents(
                contents,
                prior_obj=prior_obj,
                reason=reason,
            )

    def _update_braille_caret_position(self, obj: Atspi.Accessible) -> None:
        """Try to reposition the cursor without having to do a full update."""

        if "\ufffc" in AXText.get_all_text(obj):
            self.update_braille(obj)
            return

        super()._update_braille_caret_position(obj)

    def update_braille(
        self,
        obj: Atspi.Accessible,
        offset: int | None = None,
    ) -> None:
        """Updates the braille display to show the given object."""

        tokens = ["WEB: updating braille for", obj, "offset:", offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True, True)

        if not braille_presenter.get_presenter().use_braille():
            return

        if document_presenter.get_presenter().in_focus_mode(
            self.app,
        ) and "\ufffc" not in AXText.get_all_text(obj):
            tokens = ["WEB: updating braille in focus mode", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            super().update_braille(obj, offset=offset)
            return

        document = self.utilities.get_top_level_document_for_object(obj)
        if not document:
            tokens = ["WEB: updating braille for non-document object", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            super().update_braille(obj, offset=offset)
            return

        is_content_editable = self.utilities.is_content_editable_with_embedded_objects(obj)

        handled_reason = None
        if caret_navigator.get_navigator().last_input_event_was_navigation_command():
            handled_reason = "caret navigation command"
        elif structural_navigator.get_navigator().last_input_event_was_navigation_command():
            handled_reason = "structural navigation command"
        elif table_navigator.get_navigator().last_input_event_was_navigation_command():
            handled_reason = "table navigation command"
        elif is_content_editable:
            handled_reason = "content editable"
        elif AXUtilities.is_plain_text(document):
            handled_reason = "plain text document"
        elif input_event_manager.get_manager().last_event_was_caret_selection():
            handled_reason = "caret selection"
        elif (
            last_caret_set := AXUtilities.get_last_caret_set()
        ) is not None and last_caret_set.reason == CaretSetReason.BRAILLE_PANNING:
            handled_reason = "braille panning"

        if not handled_reason:
            tokens = ["WEB: updating braille for unhandled navigation type", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            super().update_braille(obj, offset=offset)
            return

        tokens = ["WEB: updating braille via line contents:", handled_reason, obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # TODO - JD: Getting the caret context can, by side effect, update it. This in turn
        # can prevent us from presenting table column headers when braille is enabled because
        # we think they are not "new." Commit bd877203f0 addressed that, but we need to stop
        # such side effects from happening in the first place.
        if offset is None:
            obj, offset = self.utilities.get_caret_context(document)
        if offset > 0 and is_content_editable and self.utilities.treat_as_text_object(obj):
            offset = min(offset, AXText.get_character_count(obj))

        contents = self.utilities.get_line_contents_at_offset(obj, offset)
        presentation_manager.get_manager().display_contents(contents)

    def handle_braille_pan_at_edge(
        self,
        direction: braille_presenter.PanDirection,
    ) -> bool | None:
        """Handles braille panning when the presenter reaches the edge of a line."""

        if not self.utilities.in_document_content():
            return super().handle_braille_pan_at_edge(direction)

        if direction == braille_presenter.PanDirection.LEFT:
            contents = self.utilities.get_previous_line_contents()
        else:
            contents = self.utilities.get_next_line_contents()
        if not contents:
            return False

        obj, start, _end, _string = contents[0]
        self.utilities.set_caret_position(obj, start, reason=CaretSetReason.BRAILLE_PANNING)
        self.update_braille(obj, offset=start)
        presenter = braille_presenter.get_presenter()
        if direction == braille_presenter.PanDirection.LEFT:
            presenter.pan_to_end()
        else:
            presenter.pan_to_beginning()
        return True

    def locus_of_focus_changed(
        self,
        event: Atspi.Event | None,
        old_focus: Atspi.Accessible | None,
        new_focus: Atspi.Accessible | None,
    ) -> bool:
        """Handles changes of focus of interest. Returns True if this script did all needed work."""

        tokens = ["WEB: Focus changing from", old_focus, "to", new_focus]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if new_focus and not AXObject.is_valid(new_focus):
            return True

        if new_focus and AXObject.is_dead(new_focus):
            return True

        document = self.utilities.get_top_level_document_for_object(new_focus)
        if not document and self.utilities.is_document(new_focus):
            document = new_focus

        sn_navigator = structural_navigator.get_navigator()
        last_command_was_struct_nav = sn_navigator.last_input_event_was_navigation_command()

        if not document:
            msg = "WEB: Locus of focus changed to non-document obj"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            document_presenter.get_presenter().reset_find_announcement_state()

            old_document = self.utilities.get_top_level_document_for_object(old_focus)
            if not document and self.utilities.is_document(old_focus):
                old_document = old_focus

            if old_focus and not old_document:
                msg = "WEB: Not refreshing grabs because we weren't in a document before"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

            if last_command_was_struct_nav and sn_navigator.get_mode(self) == NavigationMode.GUI:
                msg = "WEB: Not refreshing grabs: Last command was GUI structural navigation"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

            reason = "locus of focus no longer in document"
            document_presenter.get_presenter().suspend_navigators(self, True, reason)
            return False

        caret_offset = 0
        if self.utilities.in_find_container(old_focus) or (
            self.utilities.is_document(new_focus)
            and old_focus == focus_manager.get_manager().get_active_window()
        ):
            context_obj, context_offset = self.utilities.get_caret_context(document)
            if context_obj and AXObject.is_valid(context_obj):
                new_focus, caret_offset = context_obj, context_offset

        if AXUtilities.is_unknown_or_redundant(new_focus):
            msg = "WEB: Event source has bogus role. Likely browser bug."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            new_focus, _offset = self.utilities.first_context(new_focus, 0)

        if self.utilities.treat_as_text_object(new_focus):
            text_offset = AXText.get_caret_offset(new_focus)
            if 0 <= text_offset <= AXText.get_character_count(new_focus):
                caret_offset = text_offset

        self.utilities.set_caret_context(new_focus, caret_offset, document)
        self.update_braille(new_focus)

        contents = None
        last_command_was_caret_nav = (
            caret_navigator.get_navigator().last_input_event_was_navigation_command()
        )
        last_command_was_struct_nav = (
            last_command_was_struct_nav
            or table_navigator.get_navigator().last_input_event_was_navigation_command()
        )
        manager = input_event_manager.get_manager()
        last_command_was_line_nav = (
            manager.last_event_was_line_navigation() and not last_command_was_caret_nav
        )

        if (
            manager.last_event_was_mouse_button()
            and event
            and event.type.startswith("object:text-caret-moved")
        ):
            msg = "WEB: Last input event was mouse button. Generating line."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            contents = self.utilities.get_line_contents_at_offset(new_focus, caret_offset)
        elif (
            self.utilities.is_content_editable_with_embedded_objects(new_focus)
            and (
                last_command_was_caret_nav
                or last_command_was_struct_nav
                or last_command_was_line_nav
            )
            and not (AXUtilities.is_table_cell(new_focus) and AXObject.get_name(new_focus))
        ):
            tokens = ["WEB: New focus", new_focus, "content editable. Generating line."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            contents = self.utilities.get_line_contents_at_offset(new_focus, caret_offset)
        elif AXUtilities.is_anchor(new_focus):
            tokens = ["WEB: New focus", new_focus, "is anchor. Generating line."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            contents = self.utilities.get_line_contents_at_offset(new_focus, 0)
        elif (
            input_event_manager.get_manager().last_event_was_page_navigation()
            and not AXUtilities.get_table(new_focus)
            and not AXUtilities.is_feed_article(new_focus)
        ):
            tokens = ["WEB: New focus", new_focus, "was scrolled to. Generating line."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            contents = self.utilities.get_line_contents_at_offset(new_focus, caret_offset)
        elif AXUtilities.is_heading(new_focus):
            tokens = ["WEB: New focus", new_focus, "is heading. Generating object."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            contents = self.utilities.get_object_contents_at_offset(new_focus, 0)
        elif self.utilities.caret_moved_to_same_page_fragment(event, old_focus):
            assert event is not None
            tokens = ["WEB: Source", event.source, "is same page fragment. Generating line."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            contents = self.utilities.get_line_contents_at_offset(new_focus, 0)
        elif (
            event
            and event.type.startswith("object:children-changed:remove")
            and document_presenter.get_presenter().is_focus_mode_widget(self, new_focus)
        ):
            tokens = [
                "WEB: New focus",
                new_focus,
                "is recovery from removed child. Generating speech.",
            ]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        elif last_command_was_line_nav and not AXObject.is_valid(old_focus):
            msg = "WEB: Last input event was line nav; old_focus is invalid. Generating line."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            contents = self.utilities.get_line_contents_at_offset(new_focus, caret_offset)
        elif (
            last_command_was_line_nav and event and event.type.startswith("object:children-changed")
        ):
            msg = "WEB: Last input event was line nav and children changed. Generating line."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            contents = self.utilities.get_line_contents_at_offset(new_focus, caret_offset)
        else:
            tokens = ["WEB: New focus", new_focus, "is not a special case. Generating speech."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if new_focus and AXObject.is_dead(new_focus):
            msg = "WEB: New focus has since died"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if self._get_queued_event("object:state-changed:focused", True):
                msg = "WEB: Have matching focused event. Not speaking contents"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

        presentation_manager.get_manager().interrupt_if_needed_for_focus_change(
            old_focus, new_focus, event
        )

        if contents:
            presentation_manager.get_manager().speak_contents(contents, prior_obj=old_focus)
        else:
            presentation_manager.get_manager().present_object(
                self,
                new_focus,
                generate_braille=False,
                prior_obj=old_focus,
            )

        document_presenter.get_presenter().update_mode_if_needed(self, old_focus, new_focus)
        return True

    def _on_active_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:active accessibility events."""

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not event.detail1:
            msg = "WEB: Ignoring because event source is now inactive"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_dialog_or_alert(event.source):
            msg = "WEB: Event handled: Setting locusOfFocus to event source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            focus_manager.get_manager().set_locus_of_focus(event, event.source)
            return True

        return False

    def _on_busy_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:busy accessibility events."""

        live_region_presenter.get_presenter().handle_busy_changed(self, event)

        if AXUtilities.has_no_size(event.source):
            msg = "WEB: Ignoring event from page with no size."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not AXUtilities.get_uri(event.source):
            msg = "WEB: Ignoring event from page with no URI."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        AXUtilities.clear_all_cache_now(event.source, "busy-changed event.")

        if event.detail1 and self._loading_content:
            msg = "WEB: Ignoring: Already loading document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        active_document = self.utilities.active_document()
        if active_document and active_document != event.source:
            msg = "WEB: Ignoring: Event source is not active document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        if not AXUtilities.is_document_web(event.source) and not AXUtilities.is_ancestor(
            focus,
            event.source,
            True,
        ):
            msg = "WEB: Ignoring: Not document and not something we're in"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.get_document_for_object(AXObject.get_parent(event.source)):
            msg = "WEB: Ignoring: Event source is nested document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        obj, offset = self.utilities.get_caret_context()
        if not AXObject.is_valid(obj):
            self.utilities.clear_caret_context()

        should_present = True
        mgr = speech_presenter.get_presenter()
        if mgr.get_only_speak_displayed_text():
            should_present = False
            msg = "WEB: Not presenting due to settings"
            debug.print_message(debug.LEVEL_INFO, msg, True)
        elif not (AXUtilities.is_showing(event.source) or AXUtilities.is_visible(event.source)):
            should_present = False
            msg = "WEB: Not presenting because source is not showing or visible"
            debug.print_message(debug.LEVEL_INFO, msg, True)
        elif not AXUtilities.get_uri(event.source):
            should_present = False
            msg = "WEB: Not presenting because source lacks URI"
            debug.print_message(debug.LEVEL_INFO, msg, True)
        elif (
            not event.detail1
            and document_presenter.get_presenter().in_focus_mode(self.app)
            and AXObject.is_valid(obj)
        ):
            should_present = False
            tokens = ["WEB: Not presenting due to focus mode for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        elif not mgr.use_verbose_speech():
            should_present = not event.detail1
            tokens = ["WEB: Brief verbosity set. Should present", obj, f": {should_present}"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if should_present and AXUtilities.get_uri(event.source).startswith("http"):
            if event.detail1:
                presentation_manager.get_manager().present_message(messages.PAGE_LOADING_START)
            elif AXObject.get_name(event.source):
                if not mgr.use_verbose_speech():
                    msg = AXObject.get_name(event.source)
                else:
                    msg = messages.PAGE_LOADING_END_NAMED % AXObject.get_name(event.source)
                presentation_manager.get_manager().present_message(msg)
            else:
                presentation_manager.get_manager().present_message(messages.PAGE_LOADING_END)

        self._loading_content = event.detail1
        if event.detail1:
            return True

        self.utilities.clear_cached_objects()
        if AXObject.is_dead(obj):
            obj = None

        if (
            not focus_manager.get_manager().focus_is_dead()
            and not self.utilities.in_document_content(focus)
            and AXUtilities.is_focused(focus)
        ):
            msg = "WEB: Not presenting content, focus is outside of document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if document_presenter.get_presenter().get_page_summary_on_load() and should_present:
            tokens = ["WEB: Getting page summary for", event.source]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            summary = AXUtilities.get_document_summary(event.source)
            if summary:
                presentation_manager.get_manager().present_message(summary)

        obj, offset = self.utilities.get_caret_context()
        if not AXUtilities.is_busy(event.source):
            document_presenter.get_presenter().update_mode_if_needed(self, focus, obj)

        if not obj:
            msg = "WEB: Could not get caret context"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if document_presenter.get_presenter().is_focus_mode_widget(self, obj):
            tokens = ["WEB: Setting locus of focus to focusModeWidget", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            focus_manager.get_manager().set_locus_of_focus(event, obj)
            return True

        if self.utilities.is_link(obj) and AXUtilities.is_focused(obj):
            tokens = ["WEB: Setting locus of focus to focused link", obj, ". No SayAll."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            focus_manager.get_manager().set_locus_of_focus(event, obj)
            return True

        if offset > 0:
            tokens = ["WEB: Setting locus of focus to context obj", obj, ". No SayAll"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            focus_manager.get_manager().set_locus_of_focus(event, obj)
            return True

        if not AXUtilities.is_focused(focus_manager.get_manager().get_locus_of_focus()):
            tokens = ["WEB: Setting locus of focus to context obj", obj, "(no notification)"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            focus_manager.get_manager().set_locus_of_focus(event, obj, False)

        self.update_braille(obj)
        if AXUtilities.get_document_uri_fragment(event.source):
            msg = "WEB: Not doing SayAll due to page fragment"
            debug.print_message(debug.LEVEL_INFO, msg, True)
        elif not document_presenter.get_presenter().get_say_all_on_load():
            msg = "WEB: Not doing SayAll due to sayAllOnLoad being False"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().speak_contents(
                self.utilities.get_line_contents_at_offset(obj, offset),
            )
        elif speech_manager.get_manager().get_speech_is_enabled_and_not_muted():
            msg = "WEB: Doing SayAll"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            say_all_presenter.get_presenter().say_all(self, None)
        else:
            msg = "WEB: Not doing SayAll due to speech being disabled or muted"
            debug.print_message(debug.LEVEL_INFO, msg, True)

        return True

    def _on_caret_moved(self, event: Atspi.Event) -> bool:
        """Callback for object:text-caret-moved accessibility events."""

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.BRAILLE_PANNING:
            msg = "WEB: Ignoring caret-moved event that is a side effect of braille panning"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        document = self.utilities.get_top_level_document_for_object(event.source)
        if not document:
            if self.utilities.event_is_browser_ui_noise_deprecated(event):
                msg = "WEB: Ignoring event believed to be browser UI noise"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            if self.utilities.event_is_browser_ui_autocomplete_noise_deprecated(event):
                msg = "WEB: Ignoring event believed to be browser UI autocomplete noise"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            if self.utilities.in_find_container() and reason not in (
                TextEventReason.NAVIGATION_BY_CHARACTER,
                TextEventReason.NAVIGATION_BY_WORD,
                TextEventReason.NAVIGATION_TO_LINE_BOUNDARY,
            ):
                msg = "WEB: Event handled: Presenting find results"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                document_presenter.get_presenter().present_find_results(event.source, event.detail1)
                return True

            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        obj, offset = self.utilities.get_caret_context(document, search_if_needed=False)
        tokens = ["WEB: Context: ", obj, ", ", offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # TODO - JD: Make this a TextEventReason.
        if caret_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Event ignored: Last command was caret nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        # TODO - JD: Make this a TextEventReason.
        if structural_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Event ignored: Last command was struct nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        # TODO - JD: Make this a TextEventReason.
        if table_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Event ignored: Last command was table nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.SAY_ALL:
            msg = "WEB: Event handled: SayAll triggered"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.MOUSE_PRIMARY_BUTTON:
            if (event.source, event.detail1) == (obj, offset):
                msg = "WEB: Event is for current caret context."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            if (event.source, event.detail1) == self._last_mouse_button_context:
                msg = "WEB: Event is for last mouse button context."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            self._last_mouse_button_context = event.source, event.detail1

            msg = "WEB: Event handled: Last command was mouse button"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.utilities.set_caret_context(event.source, event.detail1)
            notify = not AXUtilities.is_entry_descendant(obj, inclusive=True)
            focus_manager.get_manager().set_locus_of_focus(event, event.source, notify, True)
            return True

        if reason == TextEventReason.FOCUS_CHANGE:
            msg = "WEB: Event ignored: Caret moved due to focus change."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not self.utilities.event_is_from_locus_of_focus_document(event):
            msg = "WEB: Event ignored: Not from locus of focus document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.UI_UPDATE:
            msg = "WEB: Event ignored: Caret moved due to UI update."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason in [TextEventReason.TYPING, TextEventReason.TYPING_ECHOABLE]:
            msg = "WEB: Event handled: Updating position due to insertion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            focus_manager.get_manager().set_last_cursor_position(event.source, event.detail1)
            return True

        if reason in (TextEventReason.DELETE, TextEventReason.BACKSPACE):
            msg = "WEB: Event handled: Updating position due to deletion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            focus_manager.get_manager().set_last_cursor_position(event.source, event.detail1)
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        i_e_manager = input_event_manager.get_manager()
        if (
            self.utilities.is_item_for_editable_combo_box(focus, event.source)
            and not i_e_manager.last_event_was_character_navigation()
            and not i_e_manager.last_event_was_line_boundary_navigation()
        ):
            msg = "WEB: Event ignored: Editable combobox noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.event_is_autocomplete_noise_deprecated(event, document):
            msg = "WEB: Event ignored: Autocomplete noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if document_presenter.get_presenter().in_focus_mode(
            self.app,
        ) and self.utilities.caret_moved_outside_active_grid(event):
            msg = "WEB: Event ignored: Caret moved outside active grid during focus mode"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.treat_event_as_spinner_value_change_deprecated(event):
            msg = "WEB: Event handled as the value-change event we wish we'd get"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.update_braille(event.source)
            self._present_caret_moved_event(event)
            return True

        if not self.utilities.treat_as_text_object(event.source) and not AXUtilities.is_editable(
            event.source,
        ):
            msg = "WEB: Event ignored: Was for non-editable object we're treating as textless"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        notify = force = handled = False

        in_focus_mode = document_presenter.get_presenter().in_focus_mode(self.app)
        if in_focus_mode:
            obj, offset = event.source, event.detail1
        else:
            obj, offset = self.utilities.first_context(event.source, event.detail1)

        if reason == TextEventReason.NAVIGATION_BY_PAGE:
            msg = "WEB: Caret moved due to scrolling."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            notify = force = handled = True

        elif self.utilities.caret_moved_to_same_page_fragment(event):
            msg = "WEB: Caret moved to fragment."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            notify = force = handled = True

        elif (
            event.source != focus
            and AXUtilities.is_editable(event.source)
            and (AXUtilities.is_focused(event.source) or not AXUtilities.is_focusable(event.source))
        ):
            msg = "WEB: Editable object is not (yet) the locus of focus."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            notify = force = handled = (
                i_e_manager.last_event_was_line_navigation()
                or i_e_manager.last_event_was_return()
                or i_e_manager.last_event_was_command()
            )

        elif reason in (
            TextEventReason.NAVIGATION_BY_CHARACTER,
            TextEventReason.NAVIGATION_BY_WORD,
        ):
            msg = f"WEB: Caret moved due to {reason}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if (
                in_focus_mode
                and not AXUtilities.is_editable(event.source)
                and (obj, offset) == focus_manager.get_manager().get_last_cursor_position()
            ):
                msg = "WEB: Event ignored. In focus mode and caret hasn't moved."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                handled = True

        tokens = ["WEB: Setting context and focus to: ", obj, ", ", offset, f", notify: {notify}"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        self.utilities.set_caret_context(obj, offset, document)
        focus_manager.get_manager().set_locus_of_focus(event, obj, notify, force)
        return handled

    def _on_checked_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:checked accessibility events."""

        msg = "WEB: This event is is handled by the toolkit or default script."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return False

    def _on_children_added(self, event: Atspi.Event) -> bool:
        """Callback for object:children-changed:add accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "children-changed event.")

        if self.utilities.event_is_browser_ui_noise_deprecated(event):
            msg = "WEB: Ignoring event believed to be browser UI noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        is_live_region = AXUtilities.is_live_region(event.source)
        document = self.utilities.get_top_level_document_for_object(event.source)
        if document and not is_live_region:
            focus = focus_manager.get_manager().get_locus_of_focus()
            if event.source == focus:
                msg = "WEB: Dumping cache: source is focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.dump_cache(document, preserve_context=True)
            elif focus_manager.get_manager().focus_is_dead():
                msg = "WEB: Dumping cache: dead focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.dump_cache(document, preserve_context=True)
            elif AXUtilities.find_ancestor(focus, lambda x: x == event.source):
                msg = "WEB: Dumping cache: source is ancestor of focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.dump_cache(document, preserve_context=True)
            else:
                msg = "WEB: Not dumping full cache"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.clear_cached_objects()
        elif is_live_region:
            msg = "WEB: Ignoring event from live region."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True
        else:
            msg = "WEB: Could not get document for event source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._loading_content:
            msg = "WEB: Ignoring because document content is being loaded."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not AXObject.is_valid(document):
            tokens = ["WEB: Ignoring because", document, "is not valid."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if AXUtilities.is_busy(document):
            tokens = ["WEB: Ignoring because", document, "is busy."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if AXUtilities.is_alert(event.any_data):
            msg = "WEB: Presenting event.any_data"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
            self.present_object(event.any_data)

            focused = AXUtilities.get_focused_object(event.any_data)
            if focused:
                notify = not self.utilities.treat_as_text_object(focused)
                tokens = ["WEB: Setting locusOfFocus and caret context to", focused]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                focus_manager.get_manager().set_locus_of_focus(event, focused, notify)
                self.utilities.set_caret_context(focused, 0)
            return True

        return False

    def _on_children_removed(self, event: Atspi.Event) -> bool:
        """Callback for object:children-changed:removed accessibility events."""

        AXUtilities.clear_all_cache_now(event.source, "children-changed event.")

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if self._loading_content:
            msg = "WEB: Ignoring because document content is being loaded."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_live_region(event.source):
            if self.utilities.handle_event_for_removed_child(event):
                msg = "WEB: Event handled for removed live-region child."
                debug.print_message(debug.LEVEL_INFO, msg, True)
            else:
                msg = "WEB: Ignoring removal from live region."
                debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        document = self.utilities.get_top_level_document_for_object(event.source)
        if document:
            focus = focus_manager.get_manager().get_locus_of_focus()
            if event.source == focus:
                msg = "WEB: Dumping cache: source is focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.dump_cache(document, preserve_context=True)
            elif focus_manager.get_manager().focus_is_dead():
                msg = "WEB: Dumping cache: dead focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.dump_cache(document, preserve_context=True)
            elif AXUtilities.find_ancestor(focus, lambda x: x == event.source):
                msg = "WEB: Dumping cache: source is ancestor of focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.dump_cache(document, preserve_context=True)
            else:
                msg = "WEB: Not dumping full cache"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.utilities.clear_cached_objects()

        if self.utilities.handle_event_for_removed_child(event):
            msg = "WEB: Event handled for removed child."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def _on_column_reordered(self, event: Atspi.Event) -> bool:
        """Callback for object:column-reordered accessibility events."""

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if event.source != AXUtilities.get_table(focus):
            msg = "WEB: focus is not in this table"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        presentation_manager.get_manager().present_message(messages.TABLE_REORDERED_COLUMNS)
        header = AXUtilities.find_ancestor_inclusive(focus, AXUtilities.is_table_header)
        msg = AXUtilities.get_presentable_sort_order_from_header(header, True)
        if msg:
            presentation_manager.get_manager().present_message(msg)
        return True

    def _on_document_load_complete(self, event: Atspi.Event) -> bool:
        """Callback for document:load-complete accessibility events."""

        if AXUtilities.has_no_size(event.source):
            msg = "WEB: Ignoring event from page with no size."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        uri = AXUtilities.get_uri(event.source)
        if not uri:
            msg = "WEB: Ignoring event from page with no URI."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if uri.startswith("moz-extension"):
            msg = f"WEB: Ignoring event from page with URI: {uri}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        AXUtilities.clear_all_cache_now(event.source, "load-complete event.")
        if self.utilities.get_document_for_object(AXObject.get_parent(event.source)):
            msg = "WEB: Ignoring: Event source is nested document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        msg = "WEB: Updating loading state and resetting live regions"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._loading_content = False
        live_region_presenter.get_presenter().reset()
        return True

    def _on_document_load_stopped(self, event: Atspi.Event) -> bool:
        """Callback for document:load-stopped accessibility events."""

        if not AXUtilities.get_uri(event.source):
            msg = "WEB: Ignoring event from page with no URI."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.get_document_for_object(AXObject.get_parent(event.source)):
            msg = "WEB: Ignoring: Event source is nested document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        msg = "WEB: Updating loading state"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._loading_content = False
        return True

    def _on_document_reload(self, event: Atspi.Event) -> bool:
        """Callback for document:reload accessibility events."""

        if not AXUtilities.get_uri(event.source):
            msg = "WEB: Ignoring event from page with no URI."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.get_document_for_object(AXObject.get_parent(event.source)):
            msg = "WEB: Ignoring: Event source is nested document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        msg = "WEB: Updating loading state"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._loading_content = True
        return True

    def _on_expanded_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:expanded accessibility events."""

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        obj, offset = self.utilities.get_caret_context(search_if_needed=False)
        tokens = ["WEB: Caret context is", obj, ", ", offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if not AXObject.is_valid(obj) and event.source == focus:
            msg = "WEB: Setting caret context to event source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.utilities.set_caret_context(event.source, 0)

        return False

    def _on_focused_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:focused accessibility events."""

        if not event.detail1:
            msg = "WEB: Ignoring because event source lost focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        document = self.utilities.get_top_level_document_for_object(event.source)
        if not document:
            msg = "WEB: Could not get document for event source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if focus_manager.get_manager().in_say_all():
            msg = "WEB: Ignoring focus change during say all"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        prev_document = self.utilities.get_top_level_document_for_object(focus)
        if prev_document != document:
            tokens = ["WEB: document changed from", prev_document, "to", document]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        elif document == event.source:
            msg = "WEB: Ignoring focus change to document ancestor of focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_link(event.source) and AXUtilities.is_ancestor(focus, event.source):
            msg = "WEB: Ignoring focus change on link ancestor of focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if caret_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Event ignored: Last command was caret nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if structural_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Event ignored: Last command was struct nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if table_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Event ignored: Last command was table nav"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if (
            document_presenter.get_presenter().browse_mode_is_sticky(self.app)
            and not input_event_manager.get_manager().last_event_was_tab_navigation()
        ):
            msg = "WEB: Element claimed focus, but browse mode is sticky"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_embedded_descendant(event.source):
            if AXUtilities.is_tool_tip(event.source) and AXUtilities.is_ancestor(
                focus, event.source
            ):
                msg = "WEB: Event believed to be side effect of tooltip navigation."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            msg = "WEB: Event handled: Setting locusOfFocus to embedded descendant"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().interrupt_if_needed_for_focus_change(
                focus, event.source, event
            )

            focus_manager.get_manager().set_locus_of_focus(event, event.source)
            return True

        if AXUtilities.is_editable(event.source):
            msg = "WEB: Event source is editable"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if AXUtilities.is_dialog_or_alert(event.source):
            if AXUtilities.is_ancestor(focus, event.source, True):
                msg = "WEB: Ignoring event from ancestor of focus"
                debug.print_message(debug.LEVEL_INFO, msg, True)
            else:
                msg = "WEB: Event handled: Setting locusOfFocus to event source"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                focus_manager.get_manager().set_locus_of_focus(event, event.source)
            return True

        obj, offset = self.utilities.get_caret_context()
        tokens = ["WEB: Caret context is", obj, ", ", offset]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not AXObject.is_valid(obj) or prev_document != document:
            tokens = ["WEB: Clearing context - obj", obj, "is not valid or document changed"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            self.utilities.clear_caret_context()

            obj, offset = AXUtilities.search_for_caret_context(
                event.source, self.utilities.caret_policy, self.utilities.is_document
            )
            if obj:
                tokens = ["WEB: Updating focus and context to", obj, ", ", offset]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                focus_manager.get_manager().set_locus_of_focus(event, obj)
                self.utilities.set_caret_context(obj, offset)
            else:
                msg = "WEB: Search for caret context failed"
                debug.print_message(debug.LEVEL_INFO, msg, True)

        if not (AXUtilities.is_focusable(event.source) and AXUtilities.is_focused(event.source)):
            msg = "WEB: Event ignored: Source is not focusable or focused"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not AXUtilities.is_document(event.source):
            msg = "WEB: Deferring to other scripts for handling non-document source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not obj:
            msg = "WEB: Unable to get valid context object"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if input_event_manager.get_manager().last_event_was_page_navigation():
            msg = "WEB: Event handled: Focus changed due to scrolling"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            focus_manager.get_manager().set_locus_of_focus(event, obj)
            self.utilities.set_caret_context(obj, offset)
            return True

        if AXUtilities.is_anchor(obj):
            cause = "Context is anchor"
        elif not (self.utilities.is_link(obj) and not AXUtilities.is_focused(obj)):
            cause = "Context is not a non-focused link"
        elif self.utilities.is_child_of_current_fragment(obj):
            cause = "Context is child of current fragment"
        elif document == event.source and AXUtilities.get_document_uri_fragment(event.source):
            cause = "Document URI is fragment"
        else:
            return False

        tokens = ["WEB: Event handled: Setting locusOfFocus to", obj, "(", cause, ")"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        focus_manager.get_manager().set_locus_of_focus(event, obj)
        return True

    def _on_mouse_button(self, event: Atspi.Event) -> bool:
        """Callback for mouse:button accessibility events."""

        return False

    def _on_name_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:property-change:accessible-name events."""

        if self.utilities.event_is_browser_ui_noise_deprecated(event):
            msg = "WEB: Ignoring event believed to be browser UI noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def _on_row_reordered(self, event: Atspi.Event) -> bool:
        """Callback for object:row-reordered accessibility events."""

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if event.source != AXUtilities.get_table(focus):
            msg = "WEB: focus is not in this table"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        presentation_manager.get_manager().present_message(messages.TABLE_REORDERED_ROWS)
        header = AXUtilities.find_ancestor_inclusive(focus, AXUtilities.is_table_header)
        msg = AXUtilities.get_presentable_sort_order_from_header(header, True)
        if msg:
            presentation_manager.get_manager().present_message(msg)
        return True

    def _on_selected_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:selected accessibility events."""

        if self.utilities.event_is_browser_ui_autocomplete_noise_deprecated(event):
            msg = "WEB: Ignoring event believed to be browser UI autocomplete noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        focus = focus_manager.get_manager().get_locus_of_focus()
        if self.utilities.event_is_browser_ui_page_switch(event):
            msg = "WEB: Event believed to be browser UI page switch"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if event.detail1:
                # https://bugzilla.mozilla.org/show_bug.cgi?id=1867044
                AXObject.clear_cache(event.source, False, "Work around Gecko bug.")
                AXUtilities.clear_all_cache_now(reason=msg)
                presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
                self.present_object(event.source, prior_obj=focus)
            return True

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if focus != event.source:
            msg = "WEB: Ignoring because event source is not locusOfFocus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def _on_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:selection-changed accessibility events."""

        if self.utilities.event_is_browser_ui_autocomplete_noise_deprecated(event):
            msg = "WEB: Ignoring event believed to be browser UI autocomplete noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.event_is_browser_ui_page_switch(event):
            msg = "WEB: Ignoring event believed to be browser UI page switch"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if not self.utilities.in_document_content(focus_manager.get_manager().get_locus_of_focus()):
            msg = "WEB: Event ignored: locusOfFocus is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not self.utilities.event_is_from_locus_of_focus_document(event):
            msg = "WEB: Event ignored: Not from locus of focus document"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if AXUtilities.is_embedded_descendant(event.source):
            if document_presenter.get_presenter().in_focus_mode(self.app):
                # Because we cannot count on the app firing the right state-changed events
                # for descendants.
                AXObject.clear_cache(
                    event.source,
                    True,
                    "Workaround for missing events on descendants.",
                )
                msg = "WEB: Event source is embedded descendant and we're in focus mode"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return False

            msg = "WEB: Event source is embedded descendant and we're in browse mode"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.event_is_irrelevant_selection_changed_event(event):
            msg = "WEB: Event ignored: Irrelevant"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        obj, _offset = self.utilities.get_caret_context()
        ancestor = AXUtilities.get_common_ancestor(obj, event.source)
        if ancestor and self.utilities.is_text_block_element(ancestor):
            msg = "WEB: Ignoring: Common ancestor of context and event source is text block"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def _on_showing_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:state-changed:showing accessibility events."""

        if event.detail1 and self.utilities.is_browser_ui_alert(event.source):
            msg = "WEB: Event handled: Presenting event source"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
            self.present_object(event.source)
            return True

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        return True

    def _on_text_attributes_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:text-attributes-changed accessibility events."""

        msg = "WEB: This event is is handled by the toolkit or default script."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return False

    def _on_text_deleted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:delete accessibility events."""

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.PAGE_SWITCH:
            msg = "WEB: Deletion is believed to be due to page switch"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.LIVE_REGION_UPDATE:
            msg = "WEB: Ignoring deletion from live region"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)

            if reason == TextEventReason.UI_UPDATE:
                msg = "WEB: Ignoring event believed to be browser UI update"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            return False

        if reason == TextEventReason.SPIN_BUTTON_VALUE_CHANGE:
            msg = "WEB: Ignoring: Event believed to be spin button value change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.AUTO_DELETION_UNPRESENTABLE:
            msg = "WEB: Ignoring event believed to be irrelevant auto deletion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        msg = "WEB: Clearing content cache due to text deletion"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self.utilities.clear_content_cache()

        if reason in (TextEventReason.DELETE, TextEventReason.BACKSPACE):
            msg = "WEB: Event believed to be due to editable text deletion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if reason in [TextEventReason.TYPING, TextEventReason.TYPING_ECHOABLE]:
            msg = "WEB: Ignoring event believed to be due to text insertion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        obj, _offset = self.utilities.get_caret_context()
        if (
            obj
            and obj != event.source
            and not AXUtilities.find_ancestor(obj, lambda x: x == event.source)
        ):
            tokens = ["WEB: Ignoring event because it isn't", obj, "or its ancestor"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        if not AXObject.is_valid(obj):
            if self.utilities.is_link(obj):
                msg = "WEB: Focused link deleted. Taking no further action."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            obj, _offset = self.utilities.get_caret_context()
            if obj:
                focus_manager.get_manager().set_locus_of_focus(event, obj, notify_script=False)

        if not AXObject.is_valid(obj):
            msg = "WEB: Unable to get valid context object"
            debug.print_message(debug.LEVEL_INFO, msg, True)

        if not AXUtilities.is_editable(
            event.source,
        ) and not self.utilities.is_content_editable_with_embedded_objects(event.source):
            focus = focus_manager.get_manager().get_locus_of_focus()
            if focus in (event.source, AXObject.get_parent(event.source)):
                msg = "WEB: Non-editable source is locus of focus. Updating braille."
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self.update_braille(event.source)
            else:
                msg = "WEB: Done processing non-editable source"
                debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def _on_text_inserted(self, event: Atspi.Event) -> bool:
        """Callback for object:text-changed:insert accessibility events."""

        reason = AXUtilities.get_text_event_reason(event)
        if reason == TextEventReason.PAGE_SWITCH:
            msg = "WEB: Insertion is believed to be due to page switch"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.LIVE_REGION_UPDATE:
            msg = "WEB: Event is from live region"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            live_region_presenter.get_presenter().handle_event(self, event)
            return True

        if reason == TextEventReason.CHILDREN_CHANGE:
            msg = "WEB: Ignoring: Event believed to be due to children change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)

            if reason == TextEventReason.UI_UPDATE:
                msg = "WEB: Ignoring event believed to be browser UI update"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            return False

        if reason == TextEventReason.SPIN_BUTTON_VALUE_CHANGE:
            msg = "WEB: Ignoring: Event believed to be spin button value change"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if reason == TextEventReason.AUTO_INSERTION_PRESENTABLE:
            msg = "WEB: Ignoring: Event believed to be auto insertion"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        msg = "WEB: Clearing content cache due to text insertion"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self.utilities.clear_content_cache()

        document = self.utilities.get_top_level_document_for_object(event.source)
        if focus_manager.get_manager().focus_is_dead():
            msg = "WEB: Dumping cache: dead focus"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.utilities.dump_cache(document, preserve_context=True)

            if AXUtilities.is_focused(event.source):
                msg = "WEB: Event handled: Setting locusOfFocus to event source"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                focus_manager.get_manager().set_locus_of_focus(None, event.source, force=True)
                return True

        if not self.utilities.treat_as_text_object(event.source):
            msg = "WEB: Ignoring: Event source is not a text object"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        source_is_focus = event.source == focus_manager.get_manager().get_locus_of_focus()
        if not AXUtilities.is_editable(event.source):
            if not source_is_focus:
                msg = "WEB: Done processing non-editable, non-locusOfFocus source"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                return True

            if self.utilities.is_clickable_element(event.source):
                msg = "WEB: Event handled: Re-setting locusOfFocus to changed clickable"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                focus_manager.get_manager().set_locus_of_focus(None, event.source, force=True)
                return True

            msg = "WEB: Non-editable source is locus of focus. Updating braille."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.update_braille(event.source)
            return True

        if (
            not source_is_focus
            and AXUtilities.is_text_input(event.source)
            and AXUtilities.is_focused(event.source)
        ):
            msg = "WEB: Focused entry is not the locus of focus. Waiting for focus event."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def _on_text_selection_changed(self, event: Atspi.Event) -> bool:
        """Callback for object:text-selection-changed accessibility events."""

        _reason = AXUtilities.get_text_event_reason(event)
        if self.utilities.event_is_browser_ui_noise_deprecated(event):
            msg = "WEB: Ignoring event believed to be browser UI noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.in_find_container():
            msg = "WEB: Event handled: Presenting find results"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            document_presenter.get_presenter().present_find_results(event.source, 0)
            return True

        if not self.utilities.in_document_content(event.source):
            msg = "WEB: Event source is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        focus = focus_manager.get_manager().get_locus_of_focus()
        if not self.utilities.in_document_content(focus):
            msg = "WEB: Locus of focus is not in document content"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.event_is_autocomplete_noise_deprecated(event):
            msg = "WEB: Ignoring: Event believed to be autocomplete noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.event_is_spinner_noise_deprecated(event):
            msg = "WEB: Ignoring: Event believed to be spinner noise"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.event_is_for_non_navigable_text_object(event):
            msg = "WEB: Ignoring event for non-navigable text object"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if not self.utilities.treat_as_text_object(event.source):
            msg = "WEB: Ignoring: Event source is not a text object"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if (
            event.source != focus
            and AXUtilities.is_text_input(event.source)
            and AXUtilities.is_focused(event.source)
        ):
            msg = "WEB: Focused entry is not the locus of focus. Waiting for focus event."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if self.utilities.is_content_editable_with_embedded_objects(event.source):
            msg = "WEB: In content editable with embedded objects"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return False

        if structural_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Ignoring: Last input event was structural navigation command."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        if table_navigator.get_navigator().last_input_event_was_navigation_command():
            msg = "WEB: Ignoring: Last input event was table navigation command."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        char = AXText.get_character_at_offset(event.source)[0]
        manager = input_event_manager.get_manager()
        if (
            char == "\ufffc"
            and not manager.last_event_was_caret_selection()
            and not manager.last_event_was_command()
        ):
            msg = "WEB: Ignoring: Not selecting and event offset is at embedded object"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        return False

    def _on_window_activated(self, event: Atspi.Event) -> bool:
        """Callback for window:activate accessibility events."""

        msg = "WEB: Deferring to app/toolkit script"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return False

    def _on_window_deactivated(self, event: Atspi.Event) -> bool:
        """Callback for window:deactivate accessibility events."""

        msg = "WEB: Clearing command state"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._last_mouse_button_context = None, -1
        return False
